"""Contains the python model representations and their APIs."""

from copy import deepcopy
from dataclasses import dataclass
from enum import Enum
from fractions import Fraction
from typing import Tuple, cast
from deprecated import deprecated

from stormvogel import parametric
import copy
import math
import uuid
import warnings

Number = int | float | Fraction

DEPRECATED_NAMES = [("Choice", "Choices"), ("Branch", "Branches")]


def __getattr__(name):
    for old_name, new_name in DEPRECATED_NAMES:
        if name == old_name:
            warnings.warn(
                f"The '{old_name}' class or function is renamed '{new_name}'",
                DeprecationWarning,
                stacklevel=2,
            )
            return globals()[new_name]
    raise AttributeError(f"module {__name__} has no attribute {name}")


@dataclass
class Interval:
    """represents an interval value for interval models

    Args:
        bottom: the bottom (left) element of the interval
        top: the top (right) element of the interval
    """

    bottom: Number
    top: Number

    def __init__(self, bottom: Number, top: Number):
        self.bottom = bottom
        self.top = top

    def __getitem__(self, idx):
        if idx == 0:
            return self.bottom
        elif idx == 1:
            return self.top
        else:
            raise IndexError(
                "Intervals only have two elements (the bottom and top element)"
            )

    def __lt__(self, other):
        if (self.bottom, self.top) < (other.bottom, other.top):
            return True
        return False

    def __str__(self):
        return f"[{self.bottom},{self.top}]"


Value = Number | parametric.Parametric | Interval


def value_to_string(
    n: Value, use_fractions: bool = True, round_digits: int = 4, denom_limit: int = 1000
) -> str:
    """Convert a Value to a string."""
    if isinstance(n, (int, float)):
        if math.isinf(float(n)):
            return "inf"
        if use_fractions:
            return str(Fraction(n).limit_denominator(denom_limit))
        else:
            return str(round(float(n), round_digits))
    elif isinstance(
        n, Fraction
    ):  # In the case of Fraction, a denominator of zero would have caused an error before.
        if use_fractions:
            return str(n.limit_denominator(denom_limit))
        else:
            return str(round(float(n), round_digits))
    elif isinstance(n, parametric.Parametric):
        return str(n)
    elif isinstance(n, Interval):
        return f"[{value_to_string(n.bottom, use_fractions, round_digits, denom_limit)},{value_to_string(n.top, use_fractions, round_digits, denom_limit)}]"
    else:
        return str(n)


class ModelType(Enum):
    """The type of the model."""

    # implemented
    DTMC = 1
    MDP = 2
    CTMC = 3
    POMDP = 4
    MA = 5
    HMM = 6


@dataclass
class Observation:
    """Represents an observation of a state (for POMDPs and HMMs)

    Args:
        observation: the observation of a state as an integer
    """

    observation: int
    valuations: dict[str, int | float | bool] | None = None

    def format(self):
        # TODO: add a good valuations format for visualizations
        return str(self.observation)

    def __str__(self):
        return f"Obs: {self.valuations if self.valuations else ''} ({self.observation})"


def _to_state_id(state_or_id: "State | int") -> int:
    if isinstance(state_or_id, State):
        return state_or_id.id
    else:
        return state_or_id


@dataclass(frozen=True)
class Action:
    """Represents an action, e.g., in MDPs.
        Note that this action object is completely independent of its corresponding branch.
        Their relation is managed by Choices.
        Two actions with the same labels are considered equal.

    Args:
        label: The label of this action. Corresponds to Storm label.
    """

    label: str | None

    def __post_init__(self):
        if self.label == "":
            object.__setattr__(self, "label", None)

    def __lt__(self, other):
        if not isinstance(other, Action):
            return NotImplemented
        return str(self.label) < str(other.label)

    def __str__(self):
        return f"Action with label {self.label}"


# The empty action. Used for DTMCs and empty action transitions in mdps.
EmptyAction = Action(None)


@dataclass()
class State[ValueType: Value]:
    """Represents a state in a Model.

    Args:
        labels: The labels of this state. Corresponds to Storm labels.
        valuations: The valuations of this state. Corresponds to Storm valuations/features.
        id: The id of this state.
        model: The model this state belongs to.
        observation: the observation of this state in case the model is a POMDP or HMM.
        name: the name of this state. (unique identifier in case ids change)
    """

    labels: list[str]
    valuations: dict[str, int | float | bool]
    id: int
    model: "Model[ValueType]"
    observation: Observation | list[tuple[ValueType, Observation]] | None
    name: str

    def __init__(
        self,
        labels: list[str],
        valuations: dict[str, int | float | bool],
        id: int,
        model,
        observation: Observation | list[tuple[ValueType, Observation]] | None = None,
        name: str | None = None,
    ):
        self.model = model

        if id in self.model.states.keys():
            raise RuntimeError(
                "There is already a state with this id. Make sure the id is unique."
            )

        self.labels = labels
        self.valuations = valuations
        self.id = id
        self.observation = observation

        # names must be unique, because they are the identifier of a state in case ids are changed
        if name is None:
            # if the user does not provide a name, we try to pick a random string
            while True:
                name = uuid.uuid4().hex[:8]
                if name not in self.model.used_names:
                    break
            self.model.used_names.add(name)
            self.name = name
        else:
            if name in self.model.used_names:
                raise RuntimeError(
                    "There is already a state with this name. Make sure the name is unique."
                )
            self.model.used_names.add(name)
            self.name = name

    def add_label(self, label: str):
        """adds a new label to the state"""
        if label in self.labels:
            raise RuntimeError(f"The label {label} is already present in this state.")

        self.labels.append(label)

    def set_observation(
        self, observation: Observation | list[tuple[ValueType, Observation]]
    ) -> Observation | list[tuple[ValueType, Observation]]:
        """sets the observation for this state"""
        if self.model.type in (ModelType.POMDP, ModelType.HMM):
            self.observation = observation
            return self.observation
        else:
            raise RuntimeError("The model this state belongs to is not a pomdp")

    def get_observation(self) -> Observation | list[tuple[ValueType, Observation]]:
        """gets the observation"""
        if self.model.supports_observations():
            if self.observation is not None:
                return self.observation
            else:
                raise RuntimeError(
                    "This state does not have an observation yet. Add one with the set_observation function."
                )
        else:
            raise RuntimeError(
                "The model this state belongs to does not support observations"
            )

    def set_choice(self, choice: "Choices | ChoiceShorthand"):
        """Set the choice for this state."""
        self.model.set_choice(self, choice)

    def add_choice(self, choices: "Choices | ChoiceShorthand"):
        """Add choices to this state."""
        self.model.add_choice(self, choices)

    def has_choices(self) -> bool:
        return self.get_choices() is not None

    def nr_choices(self) -> int:
        """The number of choices in this state."""
        choices = self.get_choices()
        return len(choices) if choices is not None else 1

    @deprecated(version="0.10.0", reason="Use get_choices() instead")
    def get_choice(self) -> "Choices | None":
        return self.get_choices()

    def get_choices(self) -> "Choices | None":
        """Gets the choices of this state."""
        if self.id in self.model.choices:
            return self.model.get_choices(self)
        else:
            return None

    def add_valuation(self, variable: str, value: int | bool | float):
        """Adds a valuation to the state."""
        self.valuations[variable] = value

    def get_valuation(self, variable: str) -> int | bool | float:
        return self.valuations[variable]

    def available_actions(self) -> list[Action]:
        """returns the list of all available actions in this state"""
        if self.model.supports_actions():
            choices = self.get_choices()
            if choices is not None:
                return choices.actions()
        return [EmptyAction]

    def get_outgoing_transitions(
        self, action: Action | None = None
    ) -> list[tuple[ValueType, "State[ValueType]"]] | None:
        """gets the outgoing transitions of this state (after a specific action)"""

        choice = self.get_choices()
        assert choice is not None

        # if the model supports actions we need to provide an action
        if action and self.model.supports_actions():
            if self.id in self.model.choices:
                return choice.choice[action].branch
        elif not action and self.model.supports_actions():
            raise RuntimeError("You need to provide a specific action")
        else:
            if self.id in self.model.choices:
                return choice.choice[EmptyAction].branch

    def is_absorbing(self) -> bool:
        """returns whether the state has a nonzero transition going to another state or not"""

        # if the state has no choice it is trivially true
        choice = self.get_choices()
        if choice is None:
            return True

        # for all actions we check if the state has outgoing transitions to a different state with value != 0
        for _, branch in choice:
            for transition in branch:
                if transition[1] != self:
                    return False
        return True

    def is_initial(self):
        """Returns whether this state is initial."""
        return "init" in self.labels

    def __str__(self):
        res = f"id: {self.id}, labels: {self.labels}, valuations: {self.valuations}"
        if self.model.supports_observations() and self.observation is not None:
            res += f", observation: {self.observation}"
        return res

    def __eq__(self, other):
        if not isinstance(other, State):
            return False

        if self.id != other.id:
            return False

        if (
            self.model.supports_observations()
            and self.observation is not None
            and other.observation is not None
            and self.observation != other.observation
        ):
            return False

        return (
            sorted(self.labels) == sorted(other.labels)
            and self.valuations == other.valuations
        )

    def __lt__(self, other):
        if not isinstance(other, State):
            return NotImplemented
        return self.id < other.id


@dataclass(order=True)
class Branches[ValueType: Value]:
    """Represents branches, which is a distribution over states.

    Args:
        branch: The branch as a list of tuples representing the transitions.
            The first element is the probability value and the second element is the target state.
    """

    branch: list[tuple[ValueType, State[ValueType]]]

    def __init__(self, *args):
        if len(args) == 1 and isinstance(args[0], list):
            self.branch = args[0]
        elif len(args) == 2:
            self.branch = [(args[0], args[1])]
        else:
            raise TypeError(
                "expects either (list of (value,state) tuples) or (value, state)"
            )

    def get_successors(self) -> set[int]:
        return set([b[1].id for b in self.branch])

    def sort_states(self):
        """sorts the branch list by states"""
        self.branch.sort(key=lambda x: x[1])

    def __str__(self):
        parts = []
        for prob, state in self.branch:
            parts.append(f"{prob} -> {state}")
        return ", ".join(parts)

    def __eq__(self, other):
        if not isinstance(other, Branches):
            return False

        return sorted(self.branch) == sorted(other.branch)

    def __add__(self, other):
        return Branches(self.branch + other.branch)

    def sum_probabilities(self) -> Number:
        sum = 0
        for prob, _ in self.branch:
            if not isinstance(prob, Number):
                raise ValueError(
                    "Summing probabilities is only defined for numerical values."
                )
            sum += prob
        return sum

    def __iter__(self):
        return iter(self.branch)


class Choices[ValueType: Value]:
    """Represents a choice, which map actions to branches.
        Note that an EmptyAction may be used if we want a non-action choice.
        Note that a single Choices might correspond to multiple 'arrows'.

    Args:
        choice: The choice dictionary. For each available action, we have a branch containing the transitions.
    """

    choice: dict[Action, Branches[ValueType]]

    def __init__(self, choice: dict[Action, Branches[ValueType]]):
        # Input validation, see RuntimeError.
        if len(choice) > 1 and EmptyAction in choice:
            raise RuntimeError(
                "It is impossible to create a choice that contains more than one action, and an emtpy action"
            )
        self.choice = choice

    def actions(self) -> list[Action]:
        """Returns the actions for the choices"""
        return list(self.choice.keys())

    def __str__(self):
        parts = []
        for action, branch in self:
            if action == EmptyAction:
                parts.append(f"{branch}")
            else:
                parts.append(f"{action} => {branch}")
        return "; ".join(parts + [])

    def has_empty_action(self) -> bool:
        # Note that we don't have to deal with the corner case where there are both empty and non-empty choices. This is dealt with at __init__.
        return self.choice.keys() == {EmptyAction}

    def __eq__(self, other):
        if not isinstance(other, Choices):
            return False

        if len(self.choice) != len(other.choice):
            return False

        for action, other_action in zip(
            sorted(self.choice.keys()), sorted(other.choice.keys())
        ):
            if not (
                action == other_action and self.choice[action] == other.choice[action]
            ):
                return False

        return True

    def sum_probabilities(self, action: Action) -> Number:
        return self.choice[action].sum_probabilities()

    def is_stochastic(self, epsilon: Number) -> bool:
        """returns whether the probabilities in the branches sum to 1"""
        return all([abs(self.sum_probabilities(a) - 1) <= epsilon for a in self.choice])

    def has_zero_transition(self) -> bool:
        for _, branch in self:
            for transition in branch:
                if isinstance(transition[0], Number) and transition[0] == 0:
                    return True
        return False

    def __getitem__(self, item):
        return self.choice[item]

    def __iter__(self):
        return iter(self.choice.items())

    def __len__(self) -> int:
        return len(self.choice)


ChoiceShorthand = (
    list[tuple[Value, State]]
    | list[tuple[Value, int]]
    | list[tuple[Action, State]]
    | list[tuple[Action, int]]
    | dict[Action, list[tuple[Value, State]]]
    | dict[Action, list[tuple[Value, int]]]
)


def choice_from_shorthand[ValueType: Value](
    shorthand: ChoiceShorthand, model: "Model[ValueType]"
) -> Choices[ValueType]:
    """Get a Choice object from a ChoiceShorthand. Use for all choices in DTMCs and for empty actions in MDPs.

    There are two possible ways to define a ChoiceShorthand.
    - using only the probability and the target state (implies default action when in an MDP).
    - using only the action and the target state (implies probability=1)."""

    if isinstance(shorthand, dict):
        # We convert the shorthand so that we have states instead of ids
        converted_shorthand = dict()
        for action, branch in shorthand.items():
            converted_shorthand[action] = []
            for value, state in branch:
                if isinstance(state, int):
                    if state not in model.states:
                        raise RuntimeError(
                            "This id is not known yet. Use states instead."
                        )
                    state = model.get_state_by_id(state)
                converted_shorthand[action].append((value, state))
        shorthand = converted_shorthand

        transition_content = dict()
        for action, branch in shorthand.items():
            assert isinstance(action, Action)
            transition_content[action] = Branches(branch)
        return Choices(transition_content)
    else:
        # We convert the shorthand so that we have states instead of ids
        converted_shorthand = []
        for value_or_action, state in shorthand:
            if isinstance(state, int):
                if state not in model.states:
                    raise RuntimeError("This id is not known yet. Use states instead.")
                state = model.get_state_by_id(state)
            converted_shorthand.append((value_or_action, state))
        shorthand = converted_shorthand

        # Check the type of the first element
        first_element = shorthand[0][0]
        if isinstance(first_element, Action):
            transition_content = dict()
            for action, state in shorthand:
                assert isinstance(action, Action)
                transition_content[action] = Branches(1, state)
            return Choices(transition_content)
        elif isinstance(first_element, Value):
            return Choices(
                {EmptyAction: Branches(cast(list[tuple[Value, State]], shorthand))}
            )
        raise RuntimeError(
            f"Type of {first_element} not supported in choice {shorthand}"
        )


@dataclass()
class RewardModel[ValueType: Value]:
    """Represents a state-exit reward model.
    Args:
        name: Name of the reward model.
        model: The model this rewardmodel belongs to.
        rewards: The rewards, the keys state action pairs.
    """

    name: str
    model: "Model"
    rewards: dict[Tuple[int, Action], ValueType]
    """Rewards dict. Hashed by state id and Action.
    The function update_rewards can be called to update rewards. After this, rewards will correspond to intermediate_rewards.
    Note that in models without actions, EmptyAction will be used here."""

    def __init__(
        self, name: str, model: "Model", rewards: dict[Tuple[int, Action], ValueType]
    ):
        self.name = name
        self.rewards = rewards
        self.model = model

    def set_from_rewards_vector(self, vector: list[ValueType]) -> None:
        """Set the rewards of this model according to a (stormpy) rewards vector."""
        combined_id = 0
        self.rewards = dict()
        for id, s in self.model:
            for a in s.available_actions():
                self.rewards[s.id, a] = vector[combined_id]
                combined_id += 1

    def get_state_reward(self, state: State) -> ValueType | None:
        """Gets the reward at said state or state action pair. Return None if no reward is present."""
        if self.model.supports_actions():
            raise RuntimeError(
                "This is a model with actions. Please call the get_state_action_reward(state, action) function instead"
            )
        if (state.id, EmptyAction) in self.rewards:
            return self.rewards[state.id, EmptyAction]
        else:
            return None

    def get_state_action_reward(self, state: State, action: Action) -> ValueType | None:
        """Gets the reward at said state or state action pair. Returns None if no reward was found."""
        if self.model.supports_actions():
            if action in state.available_actions():
                if (state.id, action) in self.rewards:
                    return self.rewards[state.id, action]
                else:
                    return None
            else:
                raise RuntimeError("This action is not available in this state")
        else:
            raise RuntimeError(
                "The model this rewardmodel belongs to does not support actions"
            )

    def set_state_reward(self, state: State, value: ValueType):
        """Sets the reward at said state. If the model has actions, try to use the empty state."""
        if self.model.supports_actions():
            self.set_state_action_reward(state, EmptyAction, value)
        else:
            self.rewards[state.id, EmptyAction] = value

    def set_state_action_reward(
        self,
        state: State,
        action: Action,
        value: ValueType,
    ):
        """sets the reward at said state action pair (in case of models with actions).
        If you disable auto_update_rewards, you will need to call update_intermediate_to
        """
        if self.model.supports_actions():
            if action in state.available_actions():
                self.rewards[state.id, action] = value
            else:
                raise RuntimeError("This action is not available in this state")
        else:
            raise RuntimeError(
                "The model this rewardmodel belongs to does not support actions"
            )

    def get_reward_vector(self) -> list[ValueType]:
        """Return the rewards in a (stormpy) vector format."""
        vector = []
        for id, s in self.model:
            for a in s.available_actions():
                reward = self.rewards[s.id, a]
                if reward is None:
                    raise RuntimeError(
                        "A reward was not set. You might want to call set_unset_rewards."
                    )
                vector.append(reward)
        return vector

    def set_unset_rewards(self, value: ValueType):
        """Fills up rewards that were not set yet with the specified value.
        Use this if converting (to stormpy) doesn't work because the reward vector does not have the expected length.
        """
        for id, s in self.model:
            for a in s.available_actions():
                if (s.id, a) not in self.rewards:
                    self.rewards[s.id, a] = value

    def __lt__(self, other) -> bool:
        if not isinstance(other, RewardModel):
            return NotImplemented
        return self.name < other.name

    def __eq__(self, other) -> bool:
        if not isinstance(other, RewardModel):
            return False

        return self.name == other.name and self.rewards == other.rewards

    def __iter__(self):
        return iter(self.rewards.items())


@dataclass
class Model[ValueType: Value]:
    """Represents a model.

    Args:
        type: The model type.
        states: The states of the model. The keys are the state's ids.
        choices: The choices of this model. The keys are the state ids.
        actions: The actions of the model, if this is a model that supports actions.
        rewards: The rewardsmodels of this model.
        exit_rates: The exit rates of the model, optional if this model supports rates.
        markovian_states: list of markovian states in the case of a ma.
    """

    type: ModelType
    # Both of these are hashed by the id of the state (=rowgroup number in the matrix)
    states: dict[int, State[ValueType]]
    choices: dict[int, Choices[ValueType]]
    actions: set[Action] | None
    observations: list[Observation] | None
    rewards: list[RewardModel]
    # In ctmcs we work with rate transitions but additionally we can optionally store exit rates (hashed by id of the state)
    exit_rates: dict[int, ValueType] | None
    # In ma's we keep track of markovian states
    markovian_states: list[State[ValueType]] | None
    parametric: bool | None
    interval: bool | None

    def __init__(self, model_type: ModelType, create_initial_state: bool = True):
        self.type = model_type
        self.choices = {}
        self.states = {}
        self.rewards = []
        self.parametric = None
        self.interval = None

        # Initialize actions if those are supported by the model type
        if self.supports_actions():
            self.actions = set()
        else:
            self.actions = None

        # Initialize rates if those are supported by the model type
        if self.supports_rates():
            self.exit_rates = {}
        else:
            self.exit_rates = None

        # Initialize possible observations if those are supported by the model type
        if self.type in (ModelType.POMDP, ModelType.HMM):
            self.observations = []
        else:
            self.observations = None

        # Initialize markovian states if applicable (in the case of MA's)
        if self.type == ModelType.MA:
            self.markovian_states = []
        else:
            self.markovian_states = None

        # We also keep track of used state names
        self.used_names = set()

        # We initialize the used id counter
        self.id_counter = 0

        # We initialize the observation id counter
        self.observation_id = 0

        # Add the initial state if specified to do so
        if create_initial_state:
            self.new_state(["init"])

    def summary(self):
        """Give a short summary of the model."""
        actions_bit = (
            f"{len(self.actions)} actions, " if self.actions is not None else ""
        )
        return (
            f"{self.type} model with {len(self.get_states())} states, "
            + actions_bit
            + f"and {len(self.get_labels())} distinct labels."
        )

    @deprecated(version="0.10.0", reason="use actions instead")
    def get_actions(self) -> set[Action] | None:
        """Return the actions of the model. Returns None if actions are not supported."""
        return self.actions

    def supports_actions(self) -> bool:
        """Returns whether this model supports actions."""
        return self.type in (ModelType.MDP, ModelType.POMDP, ModelType.MA)

    def supports_rates(self) -> bool:
        """Returns whether this model supports rates."""
        return self.type in (ModelType.CTMC, ModelType.MA)

    def supports_observations(self) -> bool:
        """Returns whether this model supports observations."""
        return self.type in (ModelType.POMDP, ModelType.HMM)

    def iterate_transitions(self) -> list[tuple[ValueType, State]]:
        """iterates through all transitions in all choices of the model"""
        transitions = []
        for transition in self.choices.values():
            for action, branch in transition:
                for transition in branch:
                    transitions.append(transition)
        return transitions

    def is_interval_model(self) -> bool:
        """Returns whether this model is an interval model, i.e., containts interval values)"""
        if self.interval is None:
            # Initialize
            for transition in self.iterate_transitions():
                if isinstance(transition[0], Interval):
                    self.interval = True
                    return self.interval
            self.interval = False
        return self.interval

    def is_parametric(self) -> bool:
        """Returns whether this model contains parametric transition values"""
        if self.parametric is None:
            # Initialize
            for transition in self.iterate_transitions():
                if isinstance(transition[0], parametric.Parametric):
                    self.parametric = True
                    return self.parametric
            self.parametric = False
        return self.parametric

    def is_stochastic(self, epsilon: Number = 0.000001) -> bool | None:
        """For discrete models: Checks if all sums of outgoing transition probabilities for all states equal 1, with at most epsilon rounding error.
        For continuous models: Checks if all sums of outgoing rates sum to 0
        """

        # if the model is parametric or an interval model, it should be trivially true, as the probabilities do
        # not even sum to a constant.
        if self.is_parametric() or self.is_interval_model():
            return True

        if not self.supports_rates():
            return all(
                [
                    self.get_choices(id).is_stochastic(epsilon)
                    for id, _ in self
                    if id in self.choices
                ]
            )
        else:
            for _, state in self:
                for action in state.available_actions():
                    sum_rates = 0
                    transitions = state.get_outgoing_transitions(action)
                    assert transitions is not None
                    for transition in transitions:
                        if isinstance(transition[0], Number):
                            sum_rates += transition[0]
                    if sum_rates != 0:
                        return False

        return True

    def normalize(self):
        """Normalizes a model (for states where outgoing transition probabilities don't sum to 1, we divide each probability by the sum)"""
        if self.is_parametric() or self.is_interval_model():
            raise RuntimeError(
                "normalize method undefined for parametric or interval models"
            )

        if not self.supports_rates():
            self.add_self_loops()
            for _, state in self:
                for action in state.available_actions():
                    # we first calculate the sum
                    sum_prob = 0
                    transitions = state.get_outgoing_transitions(action)
                    assert transitions is not None
                    for tuple in transitions:
                        if isinstance(tuple[0], Number):
                            sum_prob += tuple[0]

                    # then we divide each value by the sum
                    new_transitions = []
                    for tuple in transitions:
                        if isinstance(tuple[0], Number):
                            normalized_transition = (
                                tuple[0] / sum_prob,
                                tuple[1],
                            )
                            new_transitions.append(normalized_transition)
                    self.choices[state.id].choice[action].branch = new_transitions
        else:
            # for ctmcs and mas we currently only add self loops
            self.add_self_loops()

    def get_sub_model(self, states: list[State], normalize: bool = True) -> "Model":
        """Returns a submodel of the model based on a collection of states.
        The states in the collection are the states that stay in the model."""
        sub_model = copy.deepcopy(self)
        remove = []
        for _, state in sub_model:
            if state not in states:
                remove.append(state)
        for state in remove:
            sub_model.remove_state(state, normalize=False)

        if normalize:
            sub_model.normalize()
        return sub_model

    def parameter_valuation(self, values: dict[str, Number]) -> "Model":
        """evaluates all parametric transitions with the given values and returns the induced model"""
        evaluated_model = copy.deepcopy(self)
        for state, transition in evaluated_model.choices.items():
            for action, branch in transition:
                new_branch = []
                for tup in branch:
                    if isinstance(tup[0], parametric.Parametric):
                        tup = (tup[0].evaluate(values), tup[1])
                    new_branch.append(tup)
                evaluated_model.choices[state][action].branch = new_branch

        return evaluated_model

    def get_choice_id(self, state: State[ValueType], action: Action) -> int | None:
        """
        We calculate the appropriate choice id for a given state and action.
        This choice id corresponds to the row of the matrix in the underlying matrix.
        """
        choice_id = 0
        for s in self.states.values():
            if s == state:
                for a in s.available_actions():
                    if a == action:
                        return choice_id
                    choice_id += 1
                return None
            choice_id += s.nr_choices()
        return None

    @deprecated(version="0.10.0", reason="Use get_choice_id() instead.")
    def get_state_action_id(
        self, state: State[ValueType], action: Action
    ) -> int | None:
        return self.get_choice_id(state, action)

    def get_state_action_pair(self, id: int) -> tuple[State[ValueType], Action] | None:
        """Given an state action id, we return the corresponding state action pair
        Each state action pair corresponds to an id (which also corresponds to the row of the matrix).
        """
        i = 0
        for _, s in self:
            for a in s.available_actions():
                if id == i:
                    return (s, a)
                i += 1

    def add_self_loops(self):
        """adds self loops to all states that do not have an outgoing transition"""
        for _, state in self:
            if (
                state.get_choices() is None
            ):  # TODO what if the state has a choice but it is empty?
                self.set_choice(
                    state, [(float(0) if self.supports_rates() else float(1), state)]
                )

    def add_valuation_at_remaining_states(
        self, variables: list[str] | None = None, value: int | bool | float = 0
    ):
        """sets (dummy) value to variables in all states where they don't have a value yet"""

        # we either set it at all variables or just at a given subset of variables
        if variables is None:
            v = self.get_variables()
        else:
            v = variables

        # we set the values
        for _, state in self:
            for var in v:
                if var not in state.valuations.keys():
                    state.valuations[var] = value

    def unassigned_variables(self) -> list[tuple[State[ValueType], str]]:
        """we return a list of tuples of state variable pairs that are unassigned"""
        variables = self.get_variables()

        # we check all variables in all states
        unassigned = []
        for _, state in self:
            for variable in variables:
                if variable not in state.valuations.keys():
                    unassigned.append((state, variable))
        return unassigned

    def all_states_outgoing_transition(self) -> bool:
        """checks if all states have a choice"""
        for _, state in self:
            if state.get_choices() is None:  # TODO what if there is an empty choice
                return False
        return True

    def all_non_init_states_incoming_transition(self) -> bool:
        """checks if all states except the initial state have an incoming transition"""
        # Iterate over all transitions and remove all successor states as having an ingoing transition
        remaining_states = set(self.states.keys())
        for transition in self.iterate_transitions():
            remaining_states.discard(transition[1].id)
        for s in remaining_states:
            # Initial states could have no ingoing transition
            if not self.states[s].is_initial():
                return False
        return True

    def has_zero_transition(self) -> bool:
        """checks if the model has transitions with probability zero"""
        for _, choice in self.choices.items():
            if choice.has_zero_transition():
                return True
        return False

    def add_markovian_state(self, markovian_state: State):
        """adds a state to the markovian states (in case of markov automatas)"""
        if self.type == ModelType.MA and self.markovian_states is not None:
            self.markovian_states.append(markovian_state)
        else:
            raise RuntimeError("This model is not a MA")

    def set_choice(
        self, s: State, choices: Choices[ValueType] | ChoiceShorthand
    ) -> None:
        """Set the choice for a state."""
        if not isinstance(choices, Choices):
            choices = choice_from_shorthand(choices, self)

        if choices.has_zero_transition() and not self.supports_rates():
            raise RuntimeError("All transition probabilities should be nonzero.")

        if self.actions is not None and EmptyAction in choices.choice.keys():
            self.actions.add(EmptyAction)
        self.choices[s.id] = choices

    def add_choice(
        self, s: State, choices: Choices[ValueType] | ChoiceShorthand
    ) -> None:
        """Add new choices from a state to the model. If no choice currently exists, the result will be the same as set_choice."""

        if not isinstance(choices, Choices):
            choices = choice_from_shorthand(choices, self)

        if choices.has_zero_transition() and not self.supports_rates():
            raise RuntimeError("All transition probabilities should be nonzero.")

        try:
            existing_choices = self.get_choices(s)
        except KeyError:
            # Empty choices case, act like set_choice.
            self.set_choice(s, choices)
            return

        if not self.supports_actions():
            self.choices[s.id].choice[EmptyAction].branch += choices[EmptyAction].branch
        else:
            # Adding a choice is only valid if they are both empty or both non-empty.
            if not choices.has_empty_action() and existing_choices.has_empty_action():
                raise RuntimeError(
                    "You cannot add a choice with an non-empty action to a choice which has an empty action. Use set_choice instead."
                )
            if choices.has_empty_action() and not existing_choices.has_empty_action():
                raise RuntimeError(
                    "You cannot add a choice with an empty action to a choice which has no empty action. Use set_choice instead."
                )

            # Empty action case, add the branches together.
            if choices.has_empty_action():
                self.choices[s.id].choice[EmptyAction] += choices[EmptyAction]
            else:
                for action, branch in choices:
                    assert self.actions is not None
                    if action not in self.actions:
                        self.actions.add(action)
                    self.choices[s.id].choice[action] = branch

    def get_successor_states(self, state_or_id: State[ValueType] | int) -> set[int]:
        """Returns the set of successors of state_or_id."""
        result = set()
        for _, branches in self.get_choices(state_or_id):
            result |= branches.get_successors()
        return result

    def get_choices(self, state_or_id: State[ValueType] | int) -> Choices[ValueType]:
        """Get the choices at state s. Throws a KeyError if not present."""
        return self.choices[_to_state_id(state_or_id)]

    @deprecated(version="0.10.0", reason="use get_choices instead")
    def get_choice(self, state_or_id: State[ValueType] | int) -> Choices[ValueType]:
        return self.get_choices(state_or_id)

    def get_branches(self, state_or_id: State | int) -> Branches[ValueType]:
        """Get the branch at state s. Only intended for emtpy choices, otherwise a RuntimeError is thrown."""
        choice = self.get_choices(_to_state_id(state_or_id)).choice
        if EmptyAction not in choice:
            raise RuntimeError("Called get_branch on a non-empty choice.")
        return choice[EmptyAction]

    @deprecated(version="0.10.0", reason="use get_branches instead")
    def get_branch(self, state_or_id: State[ValueType] | int) -> Branches[ValueType]:
        """Get the branch at state s. Only intended for emtpy choices, otherwise a RuntimeError is thrown."""
        return self.get_branches(state_or_id)

    def get_action_with_label(self, label: str | None) -> Action | None:
        """Get the action with provided label"""
        assert self.actions is not None
        for action in self.actions:
            if action.label == label:
                return action
        return None

    def reassign_ids(self):
        """Reassigns the ids of states, choices and rates to be in order again.
        Mainly useful to keep consistent with storm."""

        warnings.warn(
            "Warning: Using this can cause problems in your code if there are existing references to states by id."
        )

        # we change the ids in the dictionaries of the model object
        self.states = {
            new_id: value
            for new_id, (old_id, value) in enumerate(sorted(self.states.items()))
        }

        self.choices = {
            new_id: value
            for new_id, (old_id, value) in enumerate(sorted(self.choices.items()))
        }

        if self.supports_rates and self.exit_rates is not None:
            self.exit_rates = {
                new_id: value
                for new_id, (old_id, value) in enumerate(
                    sorted(self.exit_rates.items())
                )
            }

        # we change the ids in the states themselves
        for index, state in enumerate(self.states.values()):
            state.id = index

        # we also reset the counter
        self.id_counter = max(self.states.keys()) + 1

    def remove_state(
        self, state: State, normalize: bool = True, reassign_ids: bool = False
    ):
        """Properly removes a state, it can optionally normalize the model and reassign ids automatically."""

        if state in self.states.values():
            # we remove the state from the transitions
            # first we remove incoming transitions of the state
            remove_actions_index = []
            for index, transition in self.choices.items():
                for action, branch in transition:
                    for index_tuple, tuple in enumerate(branch):
                        # remove the tuple if it references the state
                        if tuple[1].id == state.id:
                            self.choices[index].choice[action].branch.pop(index_tuple)

                    # if we have empty actions we need to remove those as well (later)
                    if branch.branch == []:
                        remove_actions_index.append((action, index))
            # here we remove those empty actions (this needs to happen after the other for loops)
            for action, index in remove_actions_index:
                self.choices[index].choice.pop(action)
                # if we have no actions at all anymore, delete the choice
                if self.choices[index].choice == {} and not index == state.id:
                    self.choices.pop(index)

            # we remove the choice with the corresponding id
            self.choices.pop(state.id)

            # We remove the state
            self.states.pop(state.id)

            # we remove the exit rates from the state when applicable
            if self.supports_rates and self.exit_rates is not None:
                self.exit_rates.pop(state.id)

            # we remove the state from the markovian state list when applicable
            if self.type == ModelType.MA and self.markovian_states is not None:
                if state in self.markovian_states:
                    self.markovian_states.remove(state)

            # we normalize the model if specified to do so
            if normalize:
                self.normalize()

            # we reassign the ids if specified to do so
            if reassign_ids:
                self.reassign_ids()
        else:
            raise RuntimeError("This state is not part of this model.")

    def remove_transitions_between_states(
        self, state0: State, state1: State, normalize: bool = True
    ):
        """
        Remove the transition(s) that start in state0 and go to state1.
        Only works on models that don't support actions.
        """
        if not self.supports_actions():
            for tuple in self.choices[state0.id][EmptyAction]:
                if tuple[1] == state1:
                    self.choices[state0.id].choice[EmptyAction].branch.remove(tuple)
            # if we have empty objects we need to remove those as well
            if self.choices[state0.id].choice[EmptyAction].branch == []:
                self.choices.pop(state0.id)

            if normalize:
                self.normalize()
        else:
            raise RuntimeError(
                "This method only works for models that don't support actions."
            )

    def new_action(self, label: str | None = None) -> Action:
        """Creates a new action and returns it."""
        if not self.supports_actions():
            raise RuntimeError(
                "Called new_action on a model that does not support actions"
            )
        assert self.actions is not None
        action = Action(label)
        self.actions.add(action)
        return action

    def get_action(self, name: str) -> Action:
        """Gets an existing action."""
        if not self.supports_actions():
            raise RuntimeError(
                "Called get_action on a model that does not support actions"
            )
        assert self.actions is not None
        found_action = None
        for action in self.actions:
            if action.label is not None and name in action.label:
                found_action = action
                break

        if found_action is None:
            raise RuntimeError(f"Could not find an action with name {name}")

        return found_action

    def action(self, label: str | None) -> Action:
        """New action or get action if it exists."""
        if not self.supports_actions():
            raise RuntimeError(
                "Called method action on a model that does not support actions"
            )
        assert self.actions is not None
        action = Action(label)

        if action not in self.actions:
            self.new_action(label)
        return action

    def new_observation(
        self,
        valuations: dict[str, int | float | bool] | None = None,
        observation_id: int | None = None,
    ) -> Observation:
        if not self.supports_observations():
            raise RuntimeError(
                "Called new_observation on a model that does not support observations"
            )
        assert self.observations is not None
        if observation_id is None:
            observation_id = self.observation_id
            self.observation_id += 1

        if (found_obs := self.get_observation(observation_id)) is not None:
            raise RuntimeError(
                f"An observation with id {observation_id} already exists, namely {found_obs}."
            )

        observation = Observation(observation_id, valuations)
        self.observations.append(observation)
        return observation

    def get_observation(self, observation_id: int) -> Observation | None:
        if not self.supports_observations():
            raise RuntimeError(
                "Called get_observation on a model that does not support observations"
            )
        assert self.observations is not None
        for observation in self.observations:
            if observation.observation == observation_id:
                return observation
        return None

    def observation(
        self,
        observation_id: int | None = None,
        valuations: dict[str, int | float | bool] | None = None,
    ) -> Observation:
        """New observation or get observation if it exists."""
        if not self.supports_observations():
            raise RuntimeError(
                "Called method observation on a model that does not support observations"
            )
        assert self.observations is not None

        if (
            observation_id is not None
            and (found_obs := self.get_observation(observation_id)) is not None
        ):
            if valuations is not None and found_obs.valuations != valuations:
                raise RuntimeError(
                    f"An observation with id {observation_id} already exists, namely {found_obs}, but has different valuations than the provided ones."
                )
            return found_obs

        return self.new_observation(valuations, observation_id)

    def new_state(
        self,
        labels: list[str] | str | None = None,
        valuations: dict[str, int | bool | float] | None = None,
        observation: Observation | list[tuple[ValueType, Observation]] | None = None,
        name: str | None = None,
    ) -> State[ValueType]:
        """Creates a new state and returns it."""

        # we set the id from the counter and increase it
        state_id = self.id_counter
        self.id_counter += 1

        if isinstance(labels, list):
            state = State(
                labels,
                valuations or {},
                state_id,
                self,
                observation=observation,
                name=name,
            )
        elif isinstance(labels, str):
            state = State(
                [labels],
                valuations or {},
                state_id,
                self,
                observation=observation,
                name=name,
            )
        elif labels is None:
            state = State(
                [], valuations or {}, state_id, self, observation=observation, name=name
            )
        else:
            raise RuntimeError("Unknown type for labels.")

        self.states[state_id] = state

        return state

    def has_only_number_values(self) -> bool:
        """Checks whether all transition values, observation probabilities, exit rates, markovian states in this model are numbers."""
        for choice in self.choices.values():
            for action, branch in choice:
                for value, _ in branch:
                    if not isinstance(value, Number):
                        return False

        if self.exit_rates is not None:
            for rate in self.exit_rates.values():
                if not isinstance(rate, Number):
                    return False

        if self.supports_observations() and self.observations is not None:
            for state in self.states.values():
                observation = state.observation
                if isinstance(observation, list):
                    for value, _ in observation:
                        if not isinstance(value, Number):
                            return False

        return True

    def make_observations_deterministic(self, reassign_ids: bool = False):
        """
        In case of POMDPs or HMMs, makes the observations deterministic by splitting states with
        multiple observations into multiple states with single observations.
        """
        if not self.has_only_number_values():
            raise RuntimeError("This method only works for models with Number values.")
        if not self.supports_observations():
            raise RuntimeError(
                "This method only works for models that support observations."
            )

        for state in list(self.states.values()):
            observation = state.observation
            if isinstance(observation, list):
                new_states_distribution: list[tuple[ValueType, State]] = []
                new_choice = state.get_choice()

                # Create new states for each observation possible in this state
                for prob, obs in observation:
                    new_state = self.new_state(
                        state.labels,
                        state.valuations,
                        obs,
                        f"{state.name} {{{str(obs.observation)}}}",
                    )
                    if new_choice is not None:
                        new_state.set_choice(deepcopy(new_choice))

                    new_states_distribution.append((prob, new_state))

                for src_state_idx, choice in self.choices.items():
                    for action, branch in choice:
                        for value, dest_state in branch:
                            if dest_state.id == state.id:
                                # we need to redirect this transition to the new states
                                # we first remove the old transition
                                self.choices[src_state_idx].choice[
                                    action
                                ].branch.remove((value, dest_state))
                                # then we add the new transitions
                                for prob, new_state in new_states_distribution:
                                    new_value = cast(Number, value) * cast(Number, prob)
                                    self.choices[src_state_idx].choice[
                                        action
                                    ].branch.append(
                                        (cast(ValueType, new_value), new_state)
                                    )

                self.remove_state(state, reassign_ids=reassign_ids)

    def get_states_with_label(self, label: str) -> list[State]:
        """Get all states with a given label."""
        # TODO: slow, not sure if that will become a problem though
        collected_states = []
        for _, state in self:
            if label in state.labels:
                collected_states.append(state)
        return collected_states

    def get_state_by_id(self, state_id: int) -> State:
        """Get a state by its id."""
        if state_id not in self.states:
            raise RuntimeError("Requested a non-existing state")
        return self.states[state_id]

    def get_state_by_name(self, state_name) -> State | None:
        """Get a state by its name."""
        names = {state.name: state for state in self.states.values()}
        if state_name not in names:
            raise RuntimeError("Requested a non-existing state")

        return names[state_name]

    def get_initial_state(self) -> State:
        """Gets the initial state (contains label "init", or has id 0)."""

        if len(self.states) == 0:
            raise RuntimeError(
                "This model does not have an initial state because it does not have any states."
            )

        init = []
        for _, state in self:
            if state.is_initial():
                init.append(state)

        if len(init) == 0:
            raise RuntimeError("This model does not have an initial state.")
        elif len(init) == 1:
            return init[0]
        else:
            raise RuntimeError(
                "This model has multiple initial states, which is not supported."
            )

    def get_ordered_labels(self) -> list[list[str]]:
        """Get all the labels of this model, ordered by id.
        IMPORTANT: If a state has no label, then a value '' is inserted!"""
        return [(s.labels if len(s.labels) > 0 else []) for s in self.get_states()]

    def get_labels(self) -> set[str]:
        """Get all labels in states of this Model."""
        collected_labels: set[str] = set()
        for _id, state in self:
            collected_labels = collected_labels | set(state.labels)
        return collected_labels

    def get_variables(self) -> set[str]:
        """gets the set of all variables present in this model (corresponding to valuations"""
        variables: set[str] = set()
        for _id, state in self.states.items():
            variables = variables | set(state.valuations.keys())
        return variables

    def get_default_rewards(self) -> RewardModel:
        """Gets the default reward model, throws a RuntimeError if there is none."""
        if len(self.rewards) == 0:
            raise RuntimeError("This model has no reward models.")
        return self.rewards[0]

    def get_rewards(self, name: str) -> RewardModel:
        """Gets the reward model with the specified name. Throws a RuntimeError if said model does not exist."""
        for model in self.rewards:
            if model.name == name:
                return model
        raise RuntimeError(f"Reward model {name} not present in model.")

    def get_parameters(self) -> set[str]:
        """Returns the set of parameters of this model"""
        parameters = set()
        for transition in self.iterate_transitions():
            if isinstance(transition[0], parametric.Parametric):
                parameters = parameters.union(transition[0].get_variables())
        return parameters

    def get_states(self) -> list[State]:
        return list(self.states.values())

    def new_reward_model(self, name: str) -> RewardModel:
        """Creates a reward model with the specified name, adds it and returns it."""
        for model in self.rewards:
            if model.name == name:
                raise RuntimeError(f"Reward model {name} already present in model.")
        reward_model = RewardModel(name, self, {})
        self.rewards.append(reward_model)
        return reward_model

    def get_rate(self, state: State) -> ValueType:
        """Gets the rate of a state."""
        if not self.supports_rates() or self.exit_rates is None:
            raise RuntimeError("Cannot get a rate of a discrete-time model.")
        return self.exit_rates[state.id]

    def set_rate(self, state: State, rate: ValueType):
        """Sets the rate of a state."""
        if not self.supports_rates() or self.exit_rates is None:
            raise RuntimeError("Cannot set a rate of a discrete-time model.")
        self.exit_rates[state.id] = rate

    @deprecated(version="0.10.0", reason="use type instead.")
    def get_type(self) -> ModelType:
        """Gets the type of this model"""
        return self.type

    @property
    def nr_states(self) -> int:
        """
        Returns the number of states in this model.
        Note that not all states need to be reachable.
        """
        return len(self.states)

    @property
    def nr_choices(self) -> int:
        """
        Returns the number of choices in the model (summed over all states).
        """
        return sum(state.nr_choices() for state in self.states.values())

    def to_dot(self) -> str:
        """Generates a dot representation of this model."""
        dot = "digraph model {\n"
        for state_id, state in self:
            dot += f'{state_id} [ label = "{state_id}: {", ".join(state.labels)}" ];\n'
        for state_id, transition in self.choices.items():
            for action, branch in transition:
                if action != EmptyAction:
                    dot += f'{state_id} [ label = "", shape=point ];\n'
        for state_id, transition in self.choices.items():
            for action, branch in transition:
                if action == EmptyAction:
                    # Only draw probabilities
                    for prob, target in branch:
                        dot += f'{state_id} -> {target.id} [ label = "{prob}" ];\n'
                else:
                    # Draw actions, then probabilities
                    dot += f'{state_id} -> {state_id} [ label = "{action.label}" ];\n'
                    for prob, target in branch:
                        dot += f'{state_id} -> {target.id} [ label = "{prob}" ];\n'

        dot += "}"
        return dot

    def __str__(self) -> str:
        res = [f"{self.type}"]
        res += ["", "States:"] + [f"{state}" for (_id, state) in self]
        res += ["", "Choices:"] + [
            f"{id}: {transition}" for (id, transition) in self.choices.items()
        ]

        if self.supports_rates() and self.exit_rates is not None:
            res += ["", "Exit rates:"] + [f"{self.exit_rates}"]

        if (
            self.supports_actions()
            and self.supports_rates()
            and self.markovian_states is not None
        ):
            markovian_states = [state.id for state in self.markovian_states]
            res += ["", "Markovian states:"] + [f"{markovian_states}"]

        return "\n".join(res)

    def __eq__(self, other) -> bool:
        if not isinstance(other, Model):
            return False

        return (
            self.actions == other.actions
            and self.type == other.type
            and self.states == other.states
            and self.choices == other.choices
            and sorted(self.rewards) == sorted(other.rewards)
            and self.exit_rates == other.exit_rates
            and self.markovian_states == other.markovian_states
        )

    def __getitem__(self, state_id: int):
        return self.states[state_id]

    def __iter__(self):
        return iter(self.states.items())


def new_dtmc(create_initial_state: bool = True) -> Model:
    """Creates a DTMC."""
    return Model(ModelType.DTMC, create_initial_state)


def new_mdp(create_initial_state: bool = True) -> Model:
    """Creates an MDP."""
    return Model(ModelType.MDP, create_initial_state)


def new_ctmc(create_initial_state: bool = True) -> Model:
    """Creates a CTMC."""
    return Model(ModelType.CTMC, create_initial_state)


def new_pomdp(create_initial_state: bool = True) -> Model:
    """Creates a POMDP."""
    return Model(ModelType.POMDP, create_initial_state)


def new_hmm(create_initial_state: bool = True) -> Model:
    """Creates a HMM."""
    return Model(ModelType.HMM, create_initial_state)


def new_ma(create_initial_state: bool = True) -> Model:
    """Creates a MA."""
    return Model(ModelType.MA, create_initial_state)


def new_model(modeltype: ModelType, create_initial_state: bool = True) -> Model:
    """More general model creation function"""
    return Model(modeltype, create_initial_state)



from dataclasses import dataclass
from fractions import Fraction

Number = int | float | Fraction


@dataclass
class Polynomial:
    """
    Represents polynomials, to be used as values for parametric models.
    Polynomials are represented as a dictionary with n-dimensional tuples as keys.

    Args:
        terms: terms of the polynomial (dictionary that relates exponents to coefficients)
        variables: variables of the polynomial as a list of strings
    """

    terms: dict[tuple, float]
    variables: list[str]

    def __init__(self, variables: list[str]):
        self.terms = dict()
        self.variables = variables

    # TODO exponents may also be a single integer
    def add_term(self, exponents: tuple[int, ...], coefficient: float):
        """
        adds a term to the polynomial
        example: add_term((1,2,3,4), 5) means we add 5*(x1^1*x2^2*x3^3*x4^4)
        """

        assert isinstance(exponents, tuple)

        if exponents in self.terms.keys():
            raise RuntimeError(
                "There is already a term with these exponents in this polynomial"
            )

        if self.terms != {}:
            my_dimension = self.get_dimension()
            term_dimension = len(list(exponents))

            if my_dimension != term_dimension:
                raise RuntimeError(
                    f"The length of the exponents tuple should be: {my_dimension}"
                )
        self.terms[exponents] = float(coefficient)

    def get_dimension(self) -> int:
        """returns the number of different variables present"""
        return len(self.variables)

    def get_variables(self) -> set[str]:
        """returns the set of parameters"""
        return set(self.variables)

    def get_degree(self) -> int | None:
        """returns the degree of the polynomial"""
        if self.terms is not {}:
            largest = 0
            for term in self.terms.keys():
                current = sum(list(term))
                if current > largest:
                    largest = current
            return largest
        raise RuntimeError("A polynomial without terms does not have a degree.")

    def evaluate(self, values: dict[str, Number]) -> float:
        """evaluates the polynomial with the given values for the variables"""
        result = 0
        for exponents, coefficient in self.terms.items():
            term = coefficient
            for variable, exponent in enumerate(exponents):
                term *= values[self.variables[variable]] ** exponent
            result += term
        return result

    def __str__(self) -> str:
        s = ""
        # we iterate through each term
        for exponents, coefficient in self.terms.items():
            # we only print terms with nonzero coefficients
            if coefficient != 0:
                # we don't print coefficients that are 1
                if coefficient != 1:
                    s += f"{coefficient}*"

                # we print the variables with their corresponding powers
                # if the tuple only consists of zeroes then we are left with 1
                all_zero = True
                for variable, exponent in enumerate(exponents):
                    if exponent != 0:
                        all_zero = False
                        s += f"{self.variables[variable]}"
                        if exponent != 1:
                            s += f"^{exponent}"
                if all_zero:
                    s += "1"
                s += " + "

        return s[:-3]

    def __lt__(self, other) -> bool:
        # we first compare by degree
        self_deg = self.get_degree()
        other_deg = other.get_degree()
        if self_deg != other_deg:
            return self_deg < other_deg

        # if the degrees are equal we compare the terms lexicografically
        return sorted(self.terms.items()) < sorted(other.terms.items())

    def __eq__(self, other) -> bool:
        if not isinstance(other, Polynomial):
            return False

        return self.terms == other.terms

    def __iter__(self):
        return iter(self.terms.items())


class RationalFunction:
    """
    Represents rational functions, to be used as values for parametric models
    Rational functions are represented as a pair of polynomials.

     Args:
        numerator: Polynomial in the numerator
        denominator: Polynomial in the denominator
    """

    numerator: Polynomial
    denominator: Polynomial

    def __init__(self, numerator: Polynomial, denominator: Polynomial):
        denominator_all_zero = True
        for exponents, coefficient in denominator.terms.items():
            if coefficient != 0:
                denominator_all_zero = False

        if not denominator_all_zero:
            self.numerator = numerator
            self.denominator = denominator
        else:
            raise RuntimeError("dividision by 0 is not allowed")

    def get_dimension(self) -> int:
        """returns the number of different variables present"""
        return max(self.numerator.get_dimension(), self.denominator.get_dimension())

    def get_variables(self) -> set[str]:
        "returns the total set of variables of this rational function"
        return set(self.numerator.variables).union(set(self.denominator.variables))

    def evaluate(self, values: dict[str, Number]) -> float:
        """evaluates the rational function with the given values"""
        return self.numerator.evaluate(values) / self.denominator.evaluate(values)

    def __str__(self) -> str:
        s = "(" + str(self.numerator) + ")/(" + str(self.denominator) + ")"
        return s

    def __lt__(self, other) -> bool:
        if not isinstance(other, RationalFunction):
            return NotImplemented

        if self.numerator != other.numerator:
            return self.numerator < other.numerator
        return self.denominator < other.denominator


Parametric = Polynomial | RationalFunction



"""The stormvogel package"""

from stormvogel import layout  # NOQA
from stormvogel.layout import Layout  # NOQA

# from stormvogel.stormpy_utils.mapping import *  # NOQA
# from stormvogel.stormpy_utils.model_checking import model_checking  # NOQA
from stormvogel.model import *  # NOQA
from stormvogel.property_builder import build_property_string  # NOQA
from stormvogel.result import *  # NOQA
from stormvogel.show import *  # NOQA
from stormvogel.simulator import *  # NOQA
from stormvogel import bird  # NOQA
from stormvogel import examples  # NOQA
from stormvogel import extensions  # NOQA
from stormvogel import stormpy_utils  # NOQA
from stormvogel.visualization import JSVisualization  # NOQA
from stormvogel.stormpy_utils.model_checking import *  # NOQA
import stormvogel.communication_server  # NOQA


def is_in_notebook():
    try:
        import sys

        shell = globals().get("__IPYTHON__", None)
        if shell is not None:
            return True
        if "ipykernel" not in sys.modules:
            return False

        from IPython.core.getipython import get_ipython

        ipython = get_ipython()

        if ipython is None or "IPKernelApp" not in ipython.config:
            return False
    except ImportError:
        return False
    except AttributeError:
        return False
    return True


if is_in_notebook():
    # Import and init magic
    from stormvogel.stormpy_utils import magic as magic



import stormvogel.model
import stormvogel.result
from typing import Union

try:
    import stormpy
except ImportError:
    stormpy = None


def convert_scheduler_to_stormvogel(
    model: stormvogel.model.Model, stormpy_scheduler: "stormpy.storage.Scheduler"
):
    """Converts a stormpy scheduler to a stormvogel scheduler"""
    taken_actions = {}
    for state in model.states.values():
        av_act = state.available_actions()
        choice = stormpy_scheduler.get_choice(state.id)
        action_index = choice.get_deterministic_choice()
        taken_actions[state.id] = av_act[action_index]

    return stormvogel.result.Scheduler(model, taken_actions)


def convert_model_checking_result(
    model: stormvogel.model.Model,
    stormpy_result: Union[
        "stormpy.ExplicitQuantitativeCheckResult",
        "stormpy.ExplicitQualitativeCheckResult",
        "stormpy.ExplicitParametricQuantitativeCheckResult",
    ],
    with_scheduler: bool = True,
) -> stormvogel.result.Result | None:
    """
    Takes a model checking result from stormpy and its associated model and converts it to a stormvogel representation
    """
    assert stormpy is not None

    # we distinguish between quantitative and qualitative results
    # (determines what kind of values our result contains)
    if (
        type(stormpy_result) == stormpy.ExplicitQuantitativeCheckResult
        or type(stormpy_result) == stormpy.ExplicitParametricQuantitativeCheckResult
    ):
        values = {
            index: value for (index, value) in enumerate(stormpy_result.get_values())
        }
    elif type(stormpy_result) == stormpy.ExplicitQualitativeCheckResult:
        values = {i: stormpy_result.at(i) for i in range(0, len(model.states))}
    else:
        raise RuntimeError("Unsupported result type")

    # we check if our results and expected converted results come with a scheduler
    if stormpy_result.has_scheduler and with_scheduler:
        # we build the results object containing a converted scheduler
        stormvogel_result = stormvogel.result.Result(
            model,
            values,
            scheduler=convert_scheduler_to_stormvogel(model, stormpy_result.scheduler),
        )
    else:
        # we build the results object without a scheduler
        stormvogel_result = stormvogel.result.Result(
            model,
            values,
        )

    return stormvogel_result



import stormvogel.result
import stormvogel.model
from typing import Callable
import random
from stormvogel.model import EmptyAction


class Path:
    """
    Path object that represents a path created by a simulator on a certain model.

    Args:
        path: The path itself is a dictionary where we either store for each step a state or a state action pair,
        depending on if we are working with a dtmc or an mdp.
        model: model that the path traverses through
    """

    path: (
        list[tuple[stormvogel.model.Action, stormvogel.model.State]]
        | list[stormvogel.model.State]
    )
    model: stormvogel.model.Model

    def __init__(
        self,
        path: (
            list[tuple[stormvogel.model.Action, stormvogel.model.State]]
            | list[stormvogel.model.State]
        ),
        model: stormvogel.model.Model,
    ):
        if model.type != stormvogel.model.ModelType.MA:
            self.path = path
            self.model = model
        else:
            # TODO make the simulators work for markov automata
            raise NotImplementedError

    def get_state_in_step(self, step: int) -> stormvogel.model.State | None:
        """returns the state discovered in the given step in the path"""
        if not self.model.supports_actions():
            state = self.path[step]
            assert isinstance(state, stormvogel.model.State)
            return state
        if self.model.supports_actions():
            t = self.path[step]
            assert isinstance(t, tuple)
            action, state = t
            assert isinstance(action, stormvogel.model.Action) and isinstance(
                state, stormvogel.model.State
            )
            return state

    def get_action_in_step(self, step: int) -> stormvogel.model.Action | None:
        """returns the action discovered in the given step in the path"""
        if self.model.supports_actions():
            t = self.path[step]
            assert (
                isinstance(t, tuple)
                and isinstance(t[0], stormvogel.model.Action)
                and isinstance(t[1], stormvogel.model.State)
            )
            action = t[0]
            return action

    def get_step(
        self, step: int
    ) -> (
        tuple[stormvogel.model.Action, stormvogel.model.State] | stormvogel.model.State
    ):
        """returns the state or state action pair discovered in the given step"""
        return self.path[step]

    def to_state_action_sequence(
        self,
    ) -> list[stormvogel.model.Action | stormvogel.model.State]:
        """Convert a Path to a list containing actions and states."""
        res: list[stormvogel.model.Action | stormvogel.model.State] = [
            self.model.get_initial_state()
        ]
        for v in self.path:
            if isinstance(v, tuple):
                res += list(v)
            else:
                res.append(v)
        return res

    def __str__(self) -> str:
        path = "initial state"
        if self.model.supports_actions():
            for t in self.path:
                assert (
                    isinstance(t, tuple)
                    and isinstance(t[0], stormvogel.model.Action)
                    and isinstance(t[1], stormvogel.model.State)
                )
                path += f" --(action: {t[0].label})--> state: {t[1].id}"
        else:
            for state in self.path:
                path += f" --> state: {state.id}"
        return path

    def __eq__(self, other):
        if not isinstance(other, Path):
            return False

        if len(self.path) != len(other.path):
            return False

        if self.model.supports_actions():
            for tuple, other_tuple in zip(sorted(self.path), sorted(other.path)):
                assert not (
                    isinstance(tuple, stormvogel.model.State)
                    or isinstance(other_tuple, stormvogel.model.State)
                )
                if not (tuple[0] == other_tuple[0] or tuple[1] == other_tuple[1]):
                    return False

            return self.model == other.model
        else:
            return self.path == other.path and self.model == other.model

    def __len__(self):
        return len(self.path)


def get_action_at_state(
    state: stormvogel.model.State,
    scheduler: (
        stormvogel.result.Scheduler
        | Callable[[stormvogel.model.State], stormvogel.model.Action]
    ),
) -> stormvogel.model.Action:
    """Helper function to obtain the chosen action in a state by a scheduler."""
    assert scheduler is not None
    if isinstance(scheduler, stormvogel.result.Scheduler):
        action = scheduler.get_action_at_state(state)
    elif callable(scheduler):
        action = scheduler(state)
    else:
        raise TypeError("Must be of type Scheduler or a function")

    return action


def step(
    state: stormvogel.model.State,
    action: stormvogel.model.Action | None = None,
    seed: int | None = None,
) -> tuple[int, list[stormvogel.model.Number], list[str]]:
    """given a state, action and seed we simulate a step and return information on the state we discover"""

    # we go to the next state according to the probability distribution of the transition
    transitions = state.get_outgoing_transitions(action)
    assert transitions is not None  # what if there are no transitions?

    # we build the probability distribution
    probability_distribution = []
    for t in transitions:
        # this action is not supported for parametric and interval models
        assert isinstance(t[0], stormvogel.model.Number)
        probability_distribution.append(float(t[0]))

    # we select the next state (according to the seed)
    states = [t[1] for t in transitions]
    if seed is not None:
        rng = random.Random(seed)
        next_state = rng.choices(states, k=1, weights=probability_distribution)[0]
    else:
        next_state = random.choices(states, k=1, weights=probability_distribution)[0]

    next_state_id = next_state.id

    # we also add the rewards
    rewards = []
    if not next_state.model.supports_actions():
        for rewardmodel in next_state.model.rewards:
            rewards.append(rewardmodel.get_state_reward(next_state))
    else:
        for rewardmodel in next_state.model.rewards:
            assert action is not None
            rewards.append(rewardmodel.get_state_action_reward(state, action))

    return next_state_id, rewards, next_state.labels


def simulate_path(
    model: stormvogel.model.Model,
    steps: int = 1,
    scheduler: (
        stormvogel.result.Scheduler
        | Callable[[stormvogel.model.State], stormvogel.model.Action]
        | None
    ) = None,
    seed: int | None = None,
) -> Path:
    """
    Simulates the model and returns the path created by the process.
    Args:
        model: The stormvogel model that the simulator should run on.
        steps: The number of steps the simulator walks through the model.
        scheduler: A stormvogel scheduler to determine what actions should be taken. Random if not provided.
                    (instead of a stormvogel scheduler, a function from states to actions can also be provided.)
        seed: The seed for the function that determines for each state what the next state will be. Random seed if not provided.

    Returns a path object.
    """

    # we need to set the seed for choosing actions in case no scheduler is provided
    random.seed(seed)

    # we start adding states or state action pairs to the path
    state_id = 0
    path = []
    if not model.supports_actions():
        for i in range(steps):
            # for each step we add a state to the path
            if not model.states[state_id].is_absorbing():
                state_id, _, _ = step(
                    model.get_state_by_id(state_id),
                    seed=seed + i if seed is not None else None,
                )
                path.append(model.states[state_id])
            else:
                break
    else:
        for i in range(steps):
            # we first choose an action (randomly or according to scheduler)
            action = (
                get_action_at_state(model.get_state_by_id(state_id), scheduler)
                if scheduler
                else random.choice(model.get_state_by_id(state_id).available_actions())
            )

            # we append the next state action pair
            if not model.states[state_id].is_absorbing():
                state_id, _, _ = step(
                    model.get_state_by_id(state_id),
                    action,
                    seed=seed + i if seed is not None else None,
                )
                path.append((action, model.states[state_id]))
            else:
                break

    return Path(path, model)


def simulate(
    model: stormvogel.model.Model,
    steps: int = 1,
    runs: int = 1,
    scheduler: (
        stormvogel.result.Scheduler
        | Callable[[stormvogel.model.State], stormvogel.model.Action]
        | None
    ) = None,
    seed: int | None = None,
) -> stormvogel.model.Model | None:
    """
    Simulates the model.
    Args:
        model: The stormvogel model that the simulator should run on
        steps: The number of steps the simulator walks through the model
        runs: The number of times the model gets simulated.
        scheduler: A stormvogel scheduler to determine what actions should be taken. Random if not provided.
                    (instead of a stormvogel scheduler, a function from states to actions can also be provided.)
        seed: The seed for the function that determines for each state what the next state will be. Random seed if not provided.

    Returns the partial model discovered by all the runs of the simulator together
    """

    # we need to set the seed for choosing actions in case no scheduler is provided
    random.seed(seed)

    # we keep track of all discovered states over all runs and add them to the partial model
    # we also add the discovered rewards and actions to the partial model if present
    partial_model = stormvogel.model.new_model(model.type, create_initial_state=False)

    # we create the initial state ourselves because we want to force the name to be 0
    init = partial_model.new_state(name="0", labels="init")
    init.valuations = model.get_initial_state().valuations

    # we add each (empty) rewardmodel to the partial model
    if model.rewards:
        for index, reward in enumerate(model.rewards):
            reward_model = partial_model.new_reward_model(model.rewards[index].name)

            # we already set the rewards for the initial state/stateaction
            if model.supports_actions():
                try:
                    r = model.rewards[index].get_state_action_reward(
                        model.get_initial_state(), EmptyAction
                    )
                    assert r is not None
                    reward_model.set_state_action_reward(
                        partial_model.get_initial_state(),
                        EmptyAction,
                        r,
                    )
                except RuntimeError:
                    pass
            else:
                r = model.rewards[index].get_state_reward(model.get_initial_state())
                assert r is not None
                reward_model.set_state_reward(
                    partial_model.get_initial_state(),
                    r,
                )

    # we keep track of the following sets
    discovered_states = {0}
    discovered_transitions = set()

    # we distinguish between models with and without actions
    if not partial_model.supports_actions():
        discovered_states_before_transitions = set()
        # now we start stepping through the model for the given number of runs
        for i in range(runs):
            # we start at state 0 and we begin taking steps
            last_state_id = 0
            for j in range(steps):
                # we make a step
                state_id, reward, labels = step(
                    model.get_state_by_id(last_state_id),
                    seed=seed + i + j if seed is not None else None,
                )

                # we add to the partial model what we discovered (if new)
                if state_id not in discovered_states:
                    discovered_states.add(state_id)
                    new_state = partial_model.new_state(
                        list(labels),
                        name=str(state_id),
                        valuations=model.get_state_by_id(state_id).valuations,
                    )

                    # we add the rewards
                    for index, rewardmodel in enumerate(partial_model.rewards):
                        rewardmodel.set_state_reward(new_state, reward[index])
                else:
                    new_state = partial_model.get_state_by_name(str(state_id))

                # we also add the transitions that we travelled through, so we need to keep track of the last state
                # and of the discovered transitions so that we don't add duplicates
                if (last_state_id, state_id) not in discovered_transitions:
                    discovered_transitions.add((last_state_id, state_id))
                    choice = model.get_choices(last_state_id)

                    # we calculate the transition probability
                    probability = 0
                    for tuple in choice.choice[stormvogel.model.EmptyAction].branch:
                        if tuple[1].id == state_id:
                            assert isinstance(tuple[0], float) or isinstance(
                                tuple[0], int
                            )
                            probability += float(
                                tuple[0]
                            )  # if there are multiple transitions between the same pair of states, they collapse
                    assert new_state is not None

                    # if the starting state of the transition is known, we append the existing branch
                    # otherwise we make a new branch
                    if last_state_id in discovered_states_before_transitions:
                        discovered_states_before_transitions.add(last_state_id)
                        s = partial_model.get_state_by_name(str(last_state_id))
                        assert s is not None
                        branch = partial_model.choices[s.id].choice[
                            stormvogel.modle.EmptyAction
                        ]
                        branch.branch.append((probability, new_state))
                    else:
                        s = partial_model.get_state_by_name(str(last_state_id))
                        assert s is not None
                        s.add_choice([(probability, new_state)])

                last_state_id = state_id
    else:
        # we additionally keep track of actions
        discovered_actions = set()
        # now we start stepping through the model for the given number of runs
        for i in range(runs):
            # we start at state 0 and we begin taking steps
            last_state_id = 0
            for j in range(steps):
                # we first choose an action
                action = (
                    get_action_at_state(model.get_state_by_id(last_state_id), scheduler)
                    if scheduler
                    else random.choice(
                        model.get_state_by_id(last_state_id).available_actions()
                    )
                )

                # we add the action to the partial model (if new)
                assert partial_model.actions is not None
                if action not in partial_model.actions:
                    partial_model.new_action(action.label)

                # we get the new discovery
                state_id, reward, labels = step(
                    model.get_state_by_id(last_state_id),
                    action,
                    seed=seed + i + j if seed is not None else None,
                )

                # we add the rewards.
                for index, rewardmodel in enumerate(partial_model.rewards):
                    state = model.get_state_by_id(last_state_id)
                    rewardmodel.set_state_action_reward(state, action, reward[index])

                # we add the state to the model
                if state_id not in discovered_states:
                    discovered_states.add(state_id)
                    new_state = partial_model.new_state(
                        list(labels),
                        name=str(state_id),
                        valuations=model.get_state_by_id(state_id).valuations,
                    )
                else:
                    new_state = partial_model.get_state_by_name(str(state_id))

                # we also add the transitions that we travelled through, so we need to keep track of the last state
                # and of the discovered transitions so that we don't add duplicates
                if (last_state_id, state_id, action) not in discovered_transitions:
                    transitions = model.get_state_by_id(
                        last_state_id
                    ).get_outgoing_transitions(action)
                    discovered_transitions.add((last_state_id, state_id, action))

                    # we calculate the transition probability
                    probability = 0
                    assert transitions is not None
                    for tuple in transitions:
                        if tuple[1].id == state_id:
                            assert isinstance(tuple[0], float) or isinstance(
                                tuple[0], int
                            )
                            probability += float(
                                tuple[0]
                            )  # if there are multiple transitions between the same pair of action with next state, they collapse

                    # if the starting state of the transition action pair is known, we append the existing branch
                    # otherwise we make a new branch
                    assert new_state is not None
                    if (last_state_id, action) in discovered_actions:
                        s = partial_model.get_state_by_name(str(last_state_id))
                        assert s is not None
                        branch = partial_model.choices[s.id].choice[action]
                        branch.branch.append((probability, new_state))
                    else:
                        discovered_actions.add((last_state_id, action))
                        s = partial_model.get_state_by_name(str(last_state_id))
                        assert s is not None
                        s.add_choice({action: [(probability, new_state)]})

                last_state_id = state_id

    return partial_model



"""Communication from Javascript/HTML to IPython/Jupyter lab using a local server and requests.
Initialization by user is not recommended. It should happen automatically when creating a network.Network.

Remember that you need AT LEAST ONE AVAILABLE (and sometimes also forwarded)
PORT BETWEEN min_port AND max_port IN ORDER FOR IT TO WORK."""

import http.server
import random
import string
import threading
from typing import Any, Callable
import logging
from time import sleep
import socket
import json


def random_word(k: int) -> str:
    """Random word of length k. Used to generate event ids."""
    return "".join(random.choices(string.ascii_letters, k=k))


enable_server: bool = True
"""Disable if you don't want to use an internal communication server. Some features might break."""

localhost_address: str = "127.0.0.1"

min_port = 8889
max_port = 8905
port_range = range(min_port, max_port)
"""The range of ports that stromvogel uses. They should all be forwarded if you're on an http tunnel."""

server_port: int = 8888
"""Global variable storing the port that is being used by this process. Changes when initialize_server is called."""

events: dict[str, Callable[[str], Any]] = {}
"""Dictionary that stores currently active events, along with their function, hashed by randomly generated ids."""

server_running: bool = False
"""Global variable that is set to true when the server is running."""

server: "CommunicationServer | None" = None
"""Global variable holding the server used for this notebook. None if not initialized."""


class CommunicationServer:
    """Run a web server in the background to receive Javascript communications.
    It works by having a list of events, each with a unique id.
    The Javascript code sends a POST request to the server with the id and the data.
    The server then looks up the event with that id and calls the function associated with it.
    """

    def __init__(self, server_port: int = 8080) -> None:
        """Initialize the communication server.
        Args:
            server_port (int, optional): Defaults to 8080.
        """
        import IPython.display as ipd

        self.server_port: int = server_port
        # Define a function within javascript that posts.
        js = """
function return_id_result(url, id, data) {
        fetch(url, {
            method: 'POST',
            body: JSON.stringify({
                'id': id,
                'data': data
            })
        })
    }
"""
        ipd.display(ipd.HTML(f"<script>{js}</script>"))
        ipd.display(ipd.Javascript(js))
        # These should both do the same thing, but just in case.

        # This inner class actually runs the server.
        class InnerServer(http.server.BaseHTTPRequestHandler):
            def do_POST(self):
                """Handle POST requests.
                Call the function associated with the id in the request body.
                The argument is passed as the body of the request."""
                content_length = int(self.headers.get("Content-Length", 0))
                body = json.loads(self.rfile.read(content_length).decode("utf-8"))
                id = body["id"]
                data = json.dumps(body["data"])
                logging.info(f"Received request: {id}\n{data}")
                f = events[id]
                f(data)

            def log_message(self, format, *args):
                # Overrides default log message
                pass

        self.web_server: http.server.HTTPServer = http.server.HTTPServer(
            (localhost_address, self.server_port), InnerServer
        )
        thr = threading.Thread(target=self.__run_server)
        thr.start()

    def __run_server(self):
        """Run the server (used to put it on a thread).
        Waits 0.5 seconds and then sets global variable server_running to true.
        This is to prevent making requests too early."""
        global server_running
        try:
            logging.info(
                f"CommunicationsServer started http://{localhost_address}:{self.server_port}"
            )
            server_running = True
            self.web_server.serve_forever()
        except KeyboardInterrupt:
            pass

    def add_event(self, js: str, function: Callable[[str], Any]) -> str:
        """Add an event using some JavaScript code.
        Within your js, use the special function FUNCTION(...) to call the Python function.

        Example:
            js = "FUNCTION(37 + 42);"
            function = lambda data: print(data)
            Then the arithmetic is performed in Javascript, and the result is printed in Python.
            Note that the function is called with the result of the arithmetic as a string.


        Returns event id which can be used to remove it later.
        """
        import IPython.display as ipd

        global server_running, server
        if server is None:
            raise TimeoutError("There is no server running.")
        while not server_running:
            sleep(0.1)
            logging.debug("Waiting for server to finish booting up.")

        id = random_word(k=20)
        # Parse the RETURN.
        returning_js = js.replace(
            "FUNCTION(",
            f"return_id_result('http://127.0.0.1:{self.server_port}', '{id}', ",
        )
        ipd.display(ipd.Javascript(returning_js))
        events[id] = function
        return id

    def remove_event(self, event_id: str) -> Callable[[str], Any]:
        """Remove the event associated with this event id."""
        return events.pop(event_id)

    def result(self, js: str, timeout_seconds: float = 2.0) -> str:
        """Execute some JavaScript, then use the special function RETURN(...) to return the result.

        Example:
            js = "RETURN(37 + 42);"
            Then the function result returns "79" as a string, but the arithmatic is executed in javascript.
        """
        # We implement this by using our add_event function, and waiting for the result to be set.
        result = None  # Variable that will contain the result when received.

        def on_result(data: str):  # Called when the javascript returns a result.
            nonlocal result
            result = data

        id = self.add_event(js.replace("RETURN", "FUNCTION"), on_result)
        passed_seconds = 0
        DELTA = 0.1
        while result is None and passed_seconds < timeout_seconds:
            sleep(DELTA)
            passed_seconds += DELTA

        self.remove_event(id)
        if result is None:
            raise TimeoutError(
                f"CommunicationServer.request did not receive result in time for request {id}:\n{js}"
            )
        else:
            return result


def __warn_request():
    print(
        "Test request failed. See 'Implementation details' in docs. Disable warning by use_server=False."
    )


def __warn_server():
    print(
        "Could not start server. See 'Implementation details' in docs. Disable warning by use_server=False."
    )


def __warn_no_free_port():
    print(
        f"""No free port [{min_port, max_port}). See 'Implementation details' in docs. Disable warning by use_server=False."""
    )


def is_port_free(port: int) -> bool:
    """Return true iff the specified port is free on localhost_address.
    Thanks to StackOverflow user Rugnar
    https://stackoverflow.com/questions/2470971/fast-way-to-test-if-a-port-is-in-use-using-python
    """
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex((localhost_address, port)) != 0


def find_free_port() -> int:
    for port_no in port_range:
        if is_port_free(port_no):
            return port_no
    return -1


def initialize_server() -> CommunicationServer | None:
    """If server is None, then create a new server and store it in global variable server.
    Use the port stored in global variable server_port.
    If the server is already initialized, just return it.

    Returns:
        CommunicationServer | None: The server if successful.
    """
    import ipywidgets as widgets
    import IPython.display as ipd

    global server, server_port, enable_server
    if not enable_server:
        return None

    output = widgets.Output()
    with output:
        print(
            "Initializing communication server and sending a test message. This might take a couple of seconds."
        )
    ipd.display(output)
    try:
        if (
            server is None
        ):  # If the server is not initialized yet, try to initialize it.
            server_port = find_free_port()
            if server_port == -1:  # Could not find a free port.
                __warn_no_free_port()
                with output:
                    ipd.clear_output()
                return None
            server = CommunicationServer(server_port=server_port)
            logging.info("Succesfully initialized server.")
            try:
                server.result("RETURN('test message')")
                logging.info("Succesfully received test message.")
            except TimeoutError:  # Test request failed.
                __warn_request()
        with output:
            ipd.clear_output()
        return server
    except OSError:
        logging.warning("Server port likely taken.")
        __warn_server()
        with output:
            ipd.clear_output()



import stormvogel.model
import random


class Scheduler:
    """
    Scheduler object specifies what action to take in each state.
    All schedulers are nondeterminstic and memoryless.

    Args:
        model: model associated with the scheduler (has to support actions).
        taken_actions: for each state the action we choose in that state.
    """

    model: stormvogel.model.Model
    # taken actions are hashed by the state id
    taken_actions: dict[int, stormvogel.model.Action]

    def __init__(
        self,
        model: stormvogel.model.Model,
        taken_actions: dict[int, stormvogel.model.Action],
    ):
        self.model = model
        self.taken_actions = taken_actions

    def get_action_at_state(
        self, state: stormvogel.model.State | int
    ) -> stormvogel.model.Action:
        """returns the action in the scheduler for the given state if present in the model"""
        if isinstance(state, int):
            state = self.model.get_state_by_id(state)
        if state in self.model.states.values():
            return self.taken_actions[state.id]
        else:
            raise RuntimeError("This state is not a part of the model")

    def generate_induced_dtmc(self) -> stormvogel.model.Model | None:
        """This function resolves the nondeterminacy of the mdp and returns the scheduler induced dtmc"""
        if self.model.type == stormvogel.model.ModelType.MDP:
            induced_dtmc = stormvogel.model.new_dtmc(create_initial_state=False)

            # we initialize the reward models
            for reward_model in self.model.rewards:
                induced_dtmc.new_reward_model(reward_model.name)

            # we add all the states and transitions according to the chosen actions
            for _, state in self.model:
                induced_dtmc.new_state(labels=state.labels, valuations=state.valuations)
                action = self.get_action_at_state(state)
                choice = state.get_outgoing_transitions(action)
                assert choice is not None
                induced_dtmc.add_choice(s=state, choices=choice)

                # we also add the rewards
                for reward_model in self.model.rewards:
                    induced_reward_model = induced_dtmc.get_rewards(reward_model.name)
                    reward = reward_model.get_state_action_reward(state, action)
                    assert reward is not None
                    induced_reward_model.set_state_reward(state, reward)

            return induced_dtmc

    def __str__(self) -> str:
        return "taken actions: " + str(self.taken_actions)

    def __eq__(self, other) -> bool:
        if not isinstance(other, Scheduler):
            return False

        return self.taken_actions == other.taken_actions


def random_scheduler(model: stormvogel.model.Model) -> Scheduler:
    """Create a random scheduler for the provided model."""
    choices = {i: random.choice(s.available_actions()) for (i, s) in model}
    return Scheduler(model, taken_actions=choices)


class Result:
    """Result object represents the model checking results for a given model

    Args:
        model: stormvogel representation of the model associated with the results
        values: for each state the model checking result
        scheduler: in case the model is an mdp we can optionally store a scheduler
    """

    model: stormvogel.model.Model
    # values are hashed by the state id:
    values: dict[int, stormvogel.model.Value]
    scheduler: Scheduler | None

    def __init__(
        self,
        model: stormvogel.model.Model,
        values: dict[int, stormvogel.model.Value],
        scheduler: Scheduler | None = None,
    ):
        self.model = model
        self.values = values

        if isinstance(scheduler, Scheduler):
            self.scheduler = scheduler
        else:
            self.scheduler = None

    def get_result_of_state(
        self, state: stormvogel.model.State | int
    ) -> stormvogel.model.Value | None:
        """returns the model checking result for a given state"""
        if isinstance(state, int) and state in self.model.states.keys():
            return self.values[state]
        if (
            isinstance(state, stormvogel.model.State)
            and state in self.model.states.values()
        ):
            return self.values[state.id]
        else:
            raise RuntimeError("This state is not a part of the model")

    def __str__(self) -> str:
        return (
            "values: \n "
            + str(self.values)
            + "\n"
            + "scheduler: \n "
            + str(self.scheduler)
        )

    def maximum_result(self) -> stormvogel.model.Value:
        """Return the maximum result."""
        values = list(self.values.values())
        max_val = values[0]
        for v in values:
            if isinstance(v, stormvogel.model.Interval) or isinstance(
                v, stormvogel.parametric.Parametric
            ):
                raise RuntimeError(
                    "maximum result function does not work for interval/parametric models"
                )
            assert isinstance(v, stormvogel.model.Number)
            if v > max_val:
                max_val = v
        return max_val

    def __eq__(self, other) -> bool:
        if not isinstance(other, Result):
            return False

        return self.values == other.values and self.scheduler == other.scheduler

    def __iter__(self):
        return iter(self.values.items())



from stormvogel import model, bird


def create_knuth_yao_imc():
    # We create our interval values
    interval = model.Interval(2 / 7, 6 / 7)
    inv_interval = model.Interval(1 / 7, 5 / 7)

    # we build the knuth yao dice using the bird model builder
    def delta(s: bird.State):
        match s.s:
            case 0:
                return [(interval, bird.State(s=1)), (inv_interval, bird.State(s=2))]
            case 1:
                return [(interval, bird.State(s=3)), (inv_interval, bird.State(s=4))]
            case 2:
                return [(interval, bird.State(s=5)), (inv_interval, bird.State(s=6))]
            case 3:
                return [
                    (interval, bird.State(s=1)),
                    (inv_interval, bird.State(s=7, d=1)),
                ]
            case 4:
                return [
                    (interval, bird.State(s=7, d=2)),
                    (inv_interval, bird.State(s=7, d=3)),
                ]
            case 5:
                return [
                    (interval, bird.State(s=7, d=4)),
                    (inv_interval, bird.State(s=7, d=5)),
                ]
            case 6:
                return [
                    (interval, bird.State(s=2)),
                    (inv_interval, bird.State(s=7, d=6)),
                ]
            case 7:
                return [(1, s)]

    def labels(s: bird.State):
        if s.s == 7:
            return f"rolled{str(s.d)}"

    knuth_yao_imc = bird.build_bird(
        delta=delta,
        init=bird.State(s=0),
        labels=labels,
        modeltype=model.ModelType.DTMC,
    )

    return knuth_yao_imc


if __name__ == "__main__":
    print(create_knuth_yao_imc())



# type: ignore


def example_building_dtmcs_01():
    import stormpy

    assert stormpy is not None

    # Use the SparseMatrixBuilder for constructing the transition matrix
    builder = stormpy.SparseMatrixBuilder(
        rows=0,
        columns=0,
        entries=0,
        force_dimensions=False,
        has_custom_row_grouping=False,
    )

    # New Transition from state 0 to target state 1 with probability 0.5
    builder.add_next_value(row=0, column=1, value=0.5)
    builder.add_next_value(0, 2, 0.5)
    builder.add_next_value(1, 3, 0.5)
    builder.add_next_value(1, 4, 0.5)
    builder.add_next_value(2, 5, 0.5)
    builder.add_next_value(2, 6, 0.5)
    builder.add_next_value(3, 7, 0.5)
    builder.add_next_value(3, 1, 0.5)
    builder.add_next_value(4, 8, 0.5)
    builder.add_next_value(4, 9, 0.5)
    builder.add_next_value(5, 10, 0.5)
    builder.add_next_value(5, 11, 0.5)
    builder.add_next_value(6, 2, 0.5)
    builder.add_next_value(6, 12, 0.5)

    # Add transitions for the final states
    for s in range(7, 13):
        builder.add_next_value(s, s, 1)

    # Build matrix
    transition_matrix = builder.build()
    # print(transition_matrix)

    # State labeling
    state_labeling = stormpy.storage.StateLabeling(13)

    # Add labels
    labels = {"init", "one", "two", "three", "four", "five", "six", "done", "deadlock"}
    for label in labels:
        state_labeling.add_label(label)

    # Set label of state 0
    state_labeling.add_label_to_state("init", 0)
    # print(state_labeling.get_states("init"))

    # Set remaining labels
    state_labeling.add_label_to_state("one", 7)
    state_labeling.add_label_to_state("two", 8)
    state_labeling.add_label_to_state("three", 9)
    state_labeling.add_label_to_state("four", 10)
    state_labeling.add_label_to_state("five", 11)
    state_labeling.add_label_to_state("six", 12)

    # Set label 'done' for multiple states
    state_labeling.set_states("done", stormpy.BitVector(13, [7, 8, 9, 10, 11, 12]))
    # print(state_labeling)

    # Reward models:
    reward_models = {}
    # Create a vector representing the state-action rewards
    action_reward = [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
    reward_models["coin_flips"] = stormpy.SparseRewardModel(
        optional_state_action_reward_vector=action_reward
    )

    components = stormpy.SparseModelComponents(
        transition_matrix=transition_matrix,
        state_labeling=state_labeling,
        reward_models=reward_models,
    )

    dtmc = stormpy.storage.SparseDtmc(components)

    return dtmc


if __name__ == "__main__":
    print(example_building_dtmcs_01())



import stormvogel.model


def create_nuclear_fusion_ctmc():
    # Create a new model
    ctmc = stormvogel.model.new_ctmc()

    # hydrogen fuses into helium
    ctmc.get_state_by_id(0).set_choice([(3, ctmc.new_state("helium"))])

    # helium fuses into carbon
    ctmc.get_state_by_id(1).set_choice([(2, ctmc.new_state("carbon"))])

    # carbon fuses into iron
    ctmc.get_state_by_id(2).set_choice([(7, ctmc.new_state("iron"))])

    # supernova
    ctmc.get_state_by_id(3).set_choice([(12, ctmc.new_state("Supernova"))])

    # we add the rates which are equal to whats in the choices since the probabilities are all 1
    rates = [3, 2, 7, 12, 0]
    for i in range(5):
        ctmc.set_rate(ctmc.get_state_by_id(i), rates[i])

    # we add self loops to all states with no outgoing choices
    ctmc.add_self_loops()

    return ctmc


if __name__ == "__main__":
    # Print the resulting model in dot format.
    print(create_nuclear_fusion_ctmc().to_dot())



"""Cell magic for writing models directly in notebooks."""

import tempfile
from warnings import warn
from IPython.core.magic import register_cell_magic
from IPython.core.getipython import get_ipython

try:
    import stormpy
except ImportError:
    stormpy = None


def parse_program(line, cell, parser_function, name):
    """Parse a program using stormpy."""
    store_to_var = True
    # The line should be one word, which is the variable we write the result into
    if len(line.split()) != 1:
        warn(f"Write the program to a variable by doing %%{name} <variable>.")
        store_to_var = False

    with tempfile.NamedTemporaryFile(delete=False) as temp_file:
        temp_file.write(cell.encode("utf-8"))
        prism_filename = temp_file.name

    program = parser_function(prism_filename)
    if store_to_var:
        ipython = get_ipython()
        if ipython is not None:
            ipython.user_ns[line] = program
        else:
            raise RuntimeError("IPython not available, but variable provided")
    return program


@register_cell_magic
def prism(line, cell):
    """Prism cell magic."""
    assert stormpy is not None
    return parse_program(line, cell, stormpy.parse_prism_program, "prism")


@register_cell_magic
def jani(line, cell):
    """JANI cell magic."""
    assert stormpy is not None
    return parse_program(line, cell, stormpy.parse_jani_model, "jani")



import ipywidgets as widgets
import IPython.display as ipd


class Displayable:
    """Abstract class for displaying something."""

    def __init__(
        self,
        output: widgets.Output | None = None,
        do_display: bool = True,
        debug_output: widgets.Output = widgets.Output(),
        spam: widgets.Output = widgets.Output(),
    ) -> None:
        """Abstract class for displaying something.

        Args:
            output (widgets.Output, optional): Output window. Defaults to widgets.Output().
            do_display (bool, optional): Controls if it should display. Defaults to True.
            debug_output (widgets.Output, optional): Useful for debugging. Defaults to widgets.Output().
            spam (widgets.Output, optional): Used for side-effect displays. Defaults to widgets.Output().
        """
        if output is None:
            self.output = widgets.Output()
        else:
            self.output = output
        self.do_display: bool = do_display
        self.debug_output: widgets.Output = debug_output
        self.spam = spam
        with self.output:
            ipd.display(self.spam)

    def maybe_display_output(self):
        """Display iff do_display is enabled."""
        if self.do_display:
            ipd.display(self.output)

    def spam_side_effects(self):
        """Display self.spam and clear its output immediately."""
        with self.output:
            ipd.display(self.spam)
        with self.spam:
            ipd.clear_output()



"""Javascript/HTML code generation functions, used by the JSVisualization. Depends on having vis-network-9.1.9-patched.js, and svgcanvas.js in the stormvogel folder."""

from stormvogel.layout import PACKAGE_ROOT_DIR


# An html template on which a Network is based.
def generate_html(
    nodes_js: str, edges_js: str, options_js: str, name: str, width: int, height: int
) -> str:
    """Generate HTML that renders the network.

    Args:
        nodes_js (str): JS code that generates the nodes DataSet.
        edges_js (str): JS code that generates the edges DataSet.
        options_js (str): JS code that generates the options object.
        name (str): The name of the network. Used to create a unique variable name.
        width (int): Width of the network div in pixels.
        height (int): Height of the network div in pixels.

    We generate two scripts:
        1. One that defines the NetworkWrapper class.
        2. One that initializes a NetworkWrapper object with the specified nodes, edges, and options.
           This NetworkWrapper object is  stored as a global variable nw_{name}.
           nw_{name}.network is the visjs network.
    """

    with open(PACKAGE_ROOT_DIR + "/vis-network-9.1.9-patched.js") as f:
        visjs_library = f.read()
    with open(PACKAGE_ROOT_DIR + "/svgcanvas.js") as f:
        svg_canvas_library = f.read()
    # Note that double brackets {{ }} are used to escape characters '{' and '}'
    return f"""
<!DOCTYPE html>
<html lang="en">
  <head>
    <title>Network</title>
    <script>{visjs_library}</script>
    <script>{svg_canvas_library}</script>
    <style type="text/css">
      #{name} {{
        width: {width}px;
        height: {height}px;
        border: 1px solid lightgray;
      }}
    </style>
  </head>
  <body>
    <div id="{name}"></div>
    <script type="text/javascript">
        {generate_network_wrapper_js()}
    </script>
    <script type="text/javascript">
        {generate_init_js(nodes_js, edges_js, options_js, name)}
    </script>
  </body>
</html>
"""


def generate_init_js(nodes_js: str, edges_js: str, options_js: str, name: str) -> str:
    """Generate JS code that initializes a NetworkWrapper object, and stores it in nw_{name}.

    Args:
        nodes_js (str): JS code that generates the nodes DataSet.
        edges_js (str): JS code that generates the edges DataSet.
        options_js (str): JS code that generates the options object.
        name (str): The name of the network. Used to create a unique variable name."""
    return f"""//js
    var nodes_local = new vis.DataSet([{nodes_js}]);
    var edges_local = new vis.DataSet([{edges_js}]);
    var options_local = {options_js};
    var container_local = document.getElementById("{name}");
    var nw_{name} = new NetworkWrapper(nodes_local, edges_local, options_local, container_local)
    """


def generate_network_wrapper_js() -> str:
    """Generate JS code that defines the NetworkWrapper class."""
    return """//js
class NetworkWrapper {
  constructor(nodes, edges, options, container) {
    this.nodes = nodes;
    this.edges = edges;
    this.options = options;
    this.container = container;
    this.data = {
      nodes: nodes,
      edges: edges,
    };
    this.network = new vis.Network(container, this.data, options);
    var this_ = this; // Events will not work if you use 'this' directly :))))) No idea why.
    this_.network.setData(this_.data);

    // Set user-triggered events.
    this.network.on( 'click', function(properties) {
      var nodeId = this_.network.getNodeAt({x:properties.event.srcEvent.offsetX, y:properties.event.srcEvent.offsetY});
      this_.makeNeighborsVisible(nodeId);
    });
    this.network.on( 'doubleClick', function(properties) {
        this_.network.setData(this_.data);
    });
  }

  setNodeColor(id, color) {
    var node = this.nodes.get(id);
    node["x"] = this.network.getPosition(id)["x"];
    node["y"] = this.network.getPosition(id)["y"];
    node["color"] = color;
    this.nodes.update(node);
  }

  makeNeighborsVisible(homeId) {
    if (homeId === undefined) {
      return;
    }
    var homeNode = this.nodes.get(homeId);

    // Make outgoing nodes visible
    var nodeIds = this.network.getConnectedNodes(homeId, "to");
    for (let i = 0; i < nodeIds.length; i++) {
      var toNodeId = nodeIds[i];
      var toNode = this.nodes.get(toNodeId);
      if (toNode["hidden"]) {
        toNode["hidden"] = false;
        toNode["physics"] = true;
        toNode["x"] = this.network.getPosition(homeId)["x"];
        toNode["y"] = this.network.getPosition(homeId)["y"];
        this.nodes.update(toNode);
      }
    }
    // Make edges visible, if both of the nodes are also visible
    var edgeIds = this.network.getConnectedEdges(homeId);
    for (let i = 0; i < edgeIds.length; i++) {
      var edgeId = edgeIds[i];
      var edge = this.edges.get(edgeId);
      var fromNode = this.nodes.get(edge.from);
      var toNode = this.nodes.get(edge.to);
      if ((! fromNode["hidden"]) && (! toNode["hidden"])) {
        edge["hidden"] = false;
        edge["physics"] = true;
        this.edges.update(edge);
      }
    }
  }

  getSvg() {
    // Export the network as an svg, return the raw svg file.
    // Most of this code is from Justin Harrell's vis-svg https://github.com/justinharrell/vis-svg
    let network = this.network;
    var networkContainer = network.body.container;
    try { // For whatever strange reason, if you enable iframe, it only works with Context and it only works with C2S otherwise...
      var ctx = new C2S({width: networkContainer.clientWidth, height: networkContainer.clientHeight, embedImages: true});
    } catch(e) {
      var ctx = new Context({width: networkContainer.clientWidth, height: networkContainer.clientHeight, embedImages: true});
    }

    var canvasProto = network.canvas.__proto__;
    var currentGetContext = canvasProto.getContext;
    canvasProto.getContext = function()
    {
        return ctx;
    }
    var svgOptions = {
        nodes: {
            shapeProperties: {
                interpolation: false //so images are not scaled svg will get full image
            },
            scaling: { label: { drawThreshold : 0} },
            font:{color:'#000000'}
        },
        edges: {
            scaling: { label: { drawThreshold : 0} }
        }
    };

    network.setOptions(svgOptions);
    network.redraw();
    canvasProto.getContext = currentGetContext;
    var svg = ctx.getSerializedSvg();
    return svg;
  }
};
"""



import stormvogel.stormpy_utils.mapping as mapping
import stormvogel.stormpy_utils.convert_results as convert_results
import stormvogel.model
import stormvogel.property_builder

try:
    import stormpy
except ImportError:
    stormpy = None


def model_checking(
    model: stormvogel.model.Model, prop: str | None = None, scheduler: bool = True
) -> stormvogel.result.Result | None:
    """
    Instead of calling this function, the storm model checker can be used.
    In order to to that, convert this model using the mapping and then do storm model checking.
    After that, map it back.

    This function just performs this procedure automatically.
    """

    assert stormpy is not None

    if not model.is_stochastic():
        raise RuntimeError("Model checking only works on stochastic models.")

    # the user must provide a property string, otherwise we provide the widget for building one
    if prop:
        # we first map the model to a stormpy model
        stormpy_model = mapping.stormvogel_to_stormpy(model)

        # we perform the model checking operation
        prop = stormpy.parse_properties(prop)
        assert prop is not None
        if model.supports_actions() and scheduler:
            stormpy_result = stormpy.model_checking(
                stormpy_model, prop[0], extract_scheduler=True
            )
        else:
            stormpy_result = stormpy.model_checking(stormpy_model, prop[0])

        # to get the correct action labels, we need to convert the model back to stormvogel instead of
        # using the initial one for now. (otherwise schedulers won't work)
        stormvogel_model = mapping.stormpy_to_stormvogel(stormpy_model)

        # we convert the results
        assert stormvogel_model is not None
        stormvogel_result = convert_results.convert_model_checking_result(
            stormvogel_model, stormpy_result
        )

        return stormvogel_result
    else:
        print(
            "You have not proved a property string. You can create a simple one using this widget."
        )
        stormvogel.property_builder.build_property_string(model)
        return None


if __name__ == "__main__":
    import examples.monty_hall

    mdp = examples.monty_hall.create_monty_hall_mdp()

    rewardmodel = mdp.new_reward_model("rewardmodel")
    rewardmodel.set_from_rewards_vector(list(range(67)))
    rewardmodel2 = mdp.new_reward_model("rewardmodel2")
    rewardmodel2.set_from_rewards_vector(list(range(67)))

    print(model_checking(mdp))



import numpy as np


def example_building_ctmcs_01():
    # Building the transition matrix using numpy
    import stormpy

    transitions = np.array(
        [
            [0, 1.5, 0, 0],
            [3, 0, 1.5, 0],
            [0, 3, 0, 1.5],
            [0, 0, 3, 0],
        ],
        dtype="float64",
    )

    # Default row groups: [0,1,2,3]
    transition_matrix = stormpy.build_sparse_matrix(transitions)
    # print(transition_matrix)

    # State labeling
    state_labeling = stormpy.storage.StateLabeling(4)
    state_labels = {"empty", "init", "deadlock", "full"}
    for label in state_labels:
        state_labeling.add_label(label)

    # Adding label to states
    state_labeling.add_label_to_state("init", 0)
    state_labeling.add_label_to_state("empty", 0)
    state_labeling.add_label_to_state("full", 3)

    # Collect components
    # rate_transitions = True, because the transition values are interpreted as rates
    components = stormpy.SparseModelComponents(
        transition_matrix=transition_matrix,
        state_labeling=state_labeling,
        rate_transitions=True,
    )

    # Build the model
    ctmc = stormpy.storage.SparseCtmc(components)

    return ctmc


if __name__ == "__main__":
    example_building_ctmcs_01()



from stormvogel import bird
from stormvogel.model import ModelType


def create_debugging_mdp():
    def available_actions(s: str):
        if s == ("init", "A"):
            return ["mec", "probs"]
        return [""]

    def delta(s: str, a: bird.Action):
        if a == "mec":
            return [(1, ("X", "mec"))]
        elif s == ("X", "mec"):
            return [(1, ("Y", "mec"))]
        elif s == ("Y", "mec"):
            return [(1, ("X", "mec"))]
        elif a == "probs":
            return [(1, ("A", "E"))]
        elif s == ("A", "E"):
            return [(0.5, ("C", "D")), (0.5, ("B", "D"))]
        return [(1, s)]

    def labels(s):
        if isinstance(s, tuple):
            return list(s)
        return [s]

    return bird.build_bird(
        delta=delta,
        init=("init", "A"),
        available_actions=available_actions,
        labels=labels,
        modeltype=ModelType.MDP,
    )



"""Convert gymnasium grid environments to stormvogel models, and render them as gifs."""

from stormvogel import bird
from typing import Callable
import stormvogel.result
import stormvogel.model
import math
import imageio
import os

GRID_ACTION_LABEL_MAP = {0: "←", 1: "↓", 2: "→", 3: "↑", 4: "pickup", 5: "dropoff"}


def gymnasium_grid_to_stormvogel(
    env, action_label_map: dict[int, str] = GRID_ACTION_LABEL_MAP
):
    """Convert a FrozenLake, Taxi, or Cliffwalking gymnasium environment to an explicit stormvogel model."""
    TRANSITIONS = env.unwrapped.P
    NO_ACTIONS = env.action_space.n
    INV_MAP = {v: k for k, v in action_label_map.items()}
    ALL_ACTIONS = list(action_label_map.values())

    def action_numer_map(a: bird.Action):
        return INV_MAP[a]

    if "taxi" in env.spec.id.lower():
        # For Taxi, we need a special initial state that goes to every state. This is to account for the randomized starting position.
        init = bird.State(n=-1, done=False)
    else:
        init = bird.State(
            n=0, done=False
        )  # Otherwise, it's just 0 (Cliffwalking and FrozenLake).

    def available_actions(s: bird.State):
        if s.n == -1:
            return [""]
        return ALL_ACTIONS[:NO_ACTIONS]

    def delta(s: bird.State, a: bird.Action):
        if (
            s.n == -1
        ):  # Special taxi init state. It goes to every location that a passenger could spawn in. This should explore all states.
            # state = ((taxi_row * 5 + taxi_col) * 5 + passenger_location) * 4 + destination
            PLS = 4  # Number of passenger locations.
            return [(1 / PLS, bird.State(n=x, done=False)) for x in range(PLS)]
        trans = TRANSITIONS[s.n][action_numer_map(a)]
        return list(map(lambda x: (x[0], bird.State(n=int(x[1]), done=x[3])), trans))

    def rewards(s: bird.State, a: bird.Action) -> dict[str, stormvogel.model.Value]:
        if s.n == -1:
            return {"R": 0}
        reward = list(map(lambda x: x[2], TRANSITIONS[s.n][action_numer_map(a)]))[0]
        return {"R": reward}

    def labels(s: bird.State):
        labels = [str(s.n), str(to_coordinate(s.n, env))]
        if s.n == get_target_state(env):
            labels.append("target")
        if s.done:
            labels.append("done")
        # labels.append("always")
        return labels

    return bird.build_bird(
        delta=delta,
        init=init,
        available_actions=available_actions,
        labels=labels,
        rewards=rewards,
        modeltype=stormvogel.model.ModelType.MDP,
    )


def to_coordinate(s, env):
    """Calculate the state's coordinates. Works for FrozenLake, Cliffwalking, and Taxi"""
    num_states = env.observation_space.n
    grid_size = int(math.sqrt(num_states))
    x_target = int(s % grid_size)
    y_target = int(s // grid_size)
    return x_target, y_target


def to_state(x, y, env):
    """Calculate the state index from coordinates. Works for FrozenLake, CliffWalking, and Taxi."""
    num_states = env.observation_space.n
    grid_size = int(math.sqrt(num_states))
    return y * grid_size + x


def get_target_state(env):
    """Calculate the target state for an env. Works for FrozenLake and Cliffwalking"""
    return env.observation_space.n - 1


def to_gymnasium_scheduler(
    model: stormvogel.model.Model,
    scheduler: (
        stormvogel.result.Scheduler
        | Callable[[stormvogel.model.State], stormvogel.model.Action]
    ),
    action_label_map: dict[int, str] = GRID_ACTION_LABEL_MAP,
) -> Callable[[int], int]:
    """Convert a stormvogel scheduler to a gymnasium scheduler (for a model that was converted using gymnasium_grid_to_stormvogel).
    Args:
        model: Stormvogel model.
        scheduler: Stormvogel scheduler.
        action_label_map: Map that you also used for the call in gymnasium_grid_to_stormvogel.
    Returns a gymnasium scheduler."""
    inv_map = {v: k for k, v in action_label_map.items()}

    def gymnasium_scheduler(env_sid: int):
        # TODO change this once bird API features are a thing.
        model_state = model.get_states_with_label(str(int(env_sid)))[0]
        if isinstance(scheduler, stormvogel.result.Scheduler):
            choice = scheduler.get_action_at_state(model_state)
        elif callable(scheduler):
            choice = scheduler(model_state)  # type: ignore
        # Handle empty action (None label) - this shouldn't happen in grid environments
        # but we need to handle it for type safety
        if choice.label is None:
            raise ValueError(
                "Empty action not supported in gymnasium grid environments"
            )
        return inv_map[choice.label]

    return gymnasium_scheduler


def gymnasium_render_model_gif(
    env,
    gymnasium_scheduler: Callable[[int], int] | None = None,
    filename: str = "my_gif",
    max_length: int = 50,
    fps: int = 2,
    loop: int = 0,
) -> str:
    """Render a gymnasium model to a gif, using the gymnasium_scheduler (A map from state numbers to action numbers) to pick an action.
    Leave as None for a random action.

    Args:
        env: The gymnasium environment.
        gymnasium_scheduler: A function that takes a state number and returns an action number.
        filename: The name of the gif file to save.
        max_length: The maximum number of frames to render.
        fps: Frames per second for the gif.
        loop: Number of times to loop the gif (0 means loop forever).
    """
    frames = []  # List to store frames

    state, info = env.reset()
    for _ in range(max_length):
        frame = env.render()
        frames.append(frame)  # Store the frame

        action = (
            env.action_space.sample()
            if gymnasium_scheduler is None
            else gymnasium_scheduler(state)
        )
        state, reward, done, truncated, info = env.step(action)

        if done:
            frame = env.render()
            frames.append(frame)  # Render the last frame
            break  # Stop if the episode ends

    env.close()
    os.makedirs("gifs", exist_ok=True)
    # Save frames as a GIF
    imageio.mimsave(
        "gifs/" + filename + ".gif", frames, fps=fps, loop=loop
    )  # Adjust FPS as needed
    return "gifs/" + filename + ".gif"



import stormvogel.model
from dataclasses import dataclass
from typing import cast, Any, Callable, Sequence
import inspect
from collections import deque


@dataclass
class State:
    """bird state object. Can contain any number of any type of arguments"""

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __repr__(self):
        return f"state({self.__dict__})"

    def __hash__(self):
        return hash(str(self.__dict__))

    def __eq__(self, other):
        if isinstance(other, State):
            return self.__dict__ == other.__dict__
        return False


type Action = str


def valid_input[ValueType: stormvogel.model.Value](
    delta: Callable[[Any, Action], Any] | Callable[[Any], Any],
    init: Any,
    rewards: (
        Callable[[Any, Action], dict[str, ValueType]]
        | Callable[[Any], dict[str, ValueType]]
        | None
    ) = None,
    labels: Callable[[Any], list[str] | str | None] | None = None,
    available_actions: Callable[[Any], list[Action]] | None = None,
    observations: Callable[[Any], int | list[tuple[ValueType, int]]] | None = None,
    rates: Callable[[Any], float] | None = None,
    valuations: Callable[[Any], dict[str, float | int | bool]] | None = None,
    observation_valuations: Callable[[int], dict[str, float | int | bool]]
    | None = None,
    modeltype: stormvogel.model.ModelType = stormvogel.model.ModelType.MDP,
):
    """
    function that checks if the input for the bird model builder is valid
    it will give a runtime error if it isn't.
    """

    supports_actions = modeltype in (
        stormvogel.model.ModelType.MDP,
        stormvogel.model.ModelType.POMDP,
        stormvogel.model.ModelType.MA,
    )

    # we first check if we have an available_actions function in case our model supports actions
    if supports_actions and available_actions is None:
        raise ValueError(
            "You have to provide an available actions function for models that support actions"
        )

    # and we check if we have an observations function in case our model is a POMDP
    if (
        modeltype in (stormvogel.model.ModelType.POMDP, stormvogel.model.ModelType.HMM)
        and observations is None
    ):
        raise ValueError(
            "You have to provide an observations function for POMDPs or HMMs"
        )

    # we check if the provided functions have the right number of parameters
    if supports_actions:
        assert available_actions is not None
        sig = inspect.signature(available_actions)
        num_params = len(sig.parameters)
        if num_params != 1:
            raise ValueError(
                f"The available_actions function must take exactly one argument (state), but it takes {num_params} arguments"
            )

    sig = inspect.signature(delta)
    num_params = len(sig.parameters)
    if supports_actions:
        if num_params != 2:
            raise ValueError(
                f"Your delta function must take exactly two arguments (state and action), but it takes {num_params} arguments"
            )
    else:
        if num_params != 1:
            raise ValueError(
                f"Your delta function must take exactly one argument (state), but it takes {num_params} arguments"
            )

    if rewards is not None:
        sig = inspect.signature(rewards)
        num_params = len(sig.parameters)
        if supports_actions:
            if num_params != 2:
                raise ValueError(
                    f"The rewards function must take exactly two arguments (state, action), but it takes {num_params} arguments"
                )
        else:
            if num_params != 1:
                raise ValueError(
                    f"The rewards function must take exactly one argument (state), but it takes {num_params} arguments"
                )

    if labels is not None:
        sig = inspect.signature(labels)
        num_params = len(sig.parameters)
        if num_params != 1:
            raise ValueError(
                f"The labels function must take exactly one argument (state), but it takes {num_params} arguments"
            )

    if observations is not None:
        sig = inspect.signature(observations)
        num_params = len(sig.parameters)
        if num_params != 1:
            raise ValueError(
                f"The observations function must take exactly one argument (state), but it takes {num_params} arguments"
            )

    if valuations is not None:
        sig = inspect.signature(valuations)
        num_params = len(sig.parameters)
        if num_params != 1:
            raise ValueError(
                f"The valuations function must take exactly one argument (state), but it takes {num_params} arguments"
            )

    if observation_valuations is not None:
        sig = inspect.signature(observation_valuations)
        num_params = len(sig.parameters)
        if num_params != 1:
            raise ValueError(
                f"The observation_valuations function must take exactly one argument (observation id), but it takes {num_params} arguments"
            )


def build_bird[ValueType: stormvogel.model.Value](
    delta: Callable[
        [Any, Action], Sequence[tuple[ValueType, Any]] | Sequence[Any] | None
    ]
    | Callable[[Any], Sequence[tuple[ValueType, Any]] | Sequence[Any] | None],
    init: Any,
    rewards: (
        Callable[[Any, Action], dict[str, ValueType]]
        | Callable[[Any], dict[str, ValueType]]
        | None
    ) = None,
    labels: Callable[[Any], list[str] | str | None] | None = None,
    available_actions: Callable[[Any], list[Action]] | None = None,
    observations: Callable[[Any], int | list[tuple[ValueType, int]]] | None = None,
    rates: Callable[[Any], float] | None = None,
    valuations: Callable[[Any], dict[str, float | int | bool]] | None = None,
    observation_valuations: Callable[[int], dict[str, float | int | bool]]
    | None = None,
    modeltype: stormvogel.model.ModelType = stormvogel.model.ModelType.MDP,
    max_size: int = 10000,
) -> stormvogel.model.Model[ValueType]:
    """
    function that converts a delta function, an available_actions function an initial state and a model type
    to a stormvogel model

    this works analogous to a prism file, where the delta is the module in this case.

    (this function uses the bird classes state and action instead of the ones from stormvogel.model)
    """

    def add_new_choices(tuples, state):
        """
        helper function to add all the newly found choices and states to the model
        """
        branch = []
        if tuples is not None:
            for tup in tuples:
                # in case only a state is provided, we assume the probability is 1
                if not isinstance(tup, tuple):
                    s = tup
                    val = 1
                elif len(tup) == 2:
                    s = tup[1]
                    val = tup[0]
                else:
                    raise ValueError(
                        f"Invalid transition tuple {tup}. Expected (probability, state) or (state)."
                    )

                if s not in state_lookup:
                    new_state = model.new_state()
                    state_lookup[s] = new_state
                    branch.append((val, new_state))
                    states_to_be_visited.append(s)
                else:
                    branch.append((val, state_lookup[s]))
        else:
            # if we have no return value, we add a self loop
            branch.append((1, state_lookup[state]))
        return branch

    valid_input(
        delta,
        init,
        rewards,
        labels,
        available_actions,
        observations,
        rates,
        valuations,
        observation_valuations,
        modeltype,
    )

    # we create the model with the given type and initial state
    model = stormvogel.model.new_model(modeltype=modeltype, create_initial_state=False)
    init_state = model.new_state(labels=["init"])

    # we continue calling delta and adding new states until no states are
    # left to be visited
    states_to_be_visited = deque([init])
    # the state type needs to be hashable
    state_lookup = {init: init_state}
    while len(states_to_be_visited) > 0:
        state = states_to_be_visited.popleft()
        choice = {}

        if model.supports_actions():
            # we loop over all available actions and call the delta function for each actions
            assert available_actions is not None
            actionslist = available_actions(state)

            if actionslist is None:
                raise ValueError(
                    f"On input {state}, the available actions function does not have a return value"
                )

            if not isinstance(actionslist, list):
                raise ValueError(
                    f"On input {state}, the available actions function does not return a list. Make sure to change it to [{actionslist}]"
                )

            for action in actionslist:
                # Actions must be strings
                if not isinstance(action, str):
                    raise ValueError(
                        f"On input {state}, the available actions function returns an action that is not a string: {action}"
                    )

                # Convert empty strings to None for the empty action
                action_label = None if action == "" else action
                stormvogel_action = model.action(action_label)

                delta = cast(Callable[[Any, str], Any], delta)
                tuples = delta(state, action)

                if not isinstance(tuples, list) and tuples is not None:
                    # If the delta does not return a list, we assume it's a single transition with probability 1
                    tuples = [(1, tuples)]

                branch = add_new_choices(tuples, state)

                if branch != []:
                    choice[stormvogel_action] = stormvogel.model.Branches(branch)
        else:
            delta = cast(Callable[[Any], Any], delta)
            tuples = delta(state)

            if not isinstance(tuples, list) and tuples is not None:
                raise ValueError(
                    f"On input {state}, the delta function does not return a list. Make sure to change the format to [(<value>,<state>),...]"
                )

            branch = add_new_choices(tuples, state)

            if branch != []:
                choice[stormvogel.model.EmptyAction] = branch

        s = state_lookup[state]
        assert s is not None
        model.add_choice(
            s,
            choice,
        )

        # if at some point we discovered more than max_size states, we complain
        # if len(list(state_lookup)) % 1000 == 0:
        #     print(f"Building model is slow, discovered {len(list(state_lookup))} states...", file=sys.stderr)
        if len(state_lookup) > max_size:
            raise RuntimeError(
                f"The model you want to create has a very large amount of states (at least {max_size}). If you wish to proceed, set max_size to some larger number."
            )

    # we add the rewards
    if rewards is not None:
        if model.supports_actions():
            # we first create the right number of reward models
            assert available_actions is not None
            rewards = cast(Callable[[Any, Action], dict[str, ValueType]], rewards)
            for reward in rewards(init, available_actions(init)[0]).items():
                model.new_reward_model(reward[0])

            # we take the initial state reward to compare later
            action = available_actions(init)[0]
            initial_state_rewards = rewards(init, action)

            for state, s in state_lookup.items():
                assert available_actions is not None
                for action in available_actions(state):
                    rewarddict = rewards(state, action)

                    # we check for the rewards when the function does not return a dict object
                    # or the length is not always the same
                    if rewarddict is None:
                        raise ValueError(
                            f"On input pair {state} {action}, the rewards function does not have a return value"
                        )

                    if not isinstance(rewarddict, dict):
                        raise ValueError(
                            f"On input pair {state} {action}, the rewards function does not return a dictionary. Make sure to change it to the format {{<rewardmodel>:<reward>,...}}"
                        )
                    if rewarddict.keys() != initial_state_rewards.keys():
                        raise ValueError(
                            "Make sure that the rewards function returns a dictionary with the same keys on each return"
                        )

                    assert s is not None
                    for index, reward in enumerate(rewarddict.items()):
                        # Convert empty strings to None for the empty action
                        action_label = None if action == "" else action
                        a = model.get_action_with_label(action_label)
                        assert a is not None
                        model.rewards[index].set_state_action_reward(
                            s,
                            a,
                            reward[1],
                        )
        else:
            # we first create the right number of reward models
            rewards = cast(Callable[[Any], dict[str, ValueType]], rewards)
            for reward in rewards(init).items():
                model.new_reward_model(reward[0])

            initial_state_rewards = rewards(init)
            for state, s in state_lookup.items():
                rewarddict = rewards(state)

                # we check for the rewards when the function does not return a dict object
                # or the length is not always the same
                if rewarddict is None:
                    raise ValueError(
                        f"On input {state}, the rewards function does not have a return value"
                    )

                if not isinstance(rewarddict, dict):
                    raise ValueError(
                        f"On input {state}, the rewards function does not return a dictionary. Make sure to change it to the format {{<rewardmodel name>:<reward>,...}}"
                    )
                if rewarddict.keys() != initial_state_rewards.keys():
                    raise ValueError(
                        "Make sure that the rewards function returns a dictionary with the same keys on each return"
                    )

                s = state_lookup[state]
                assert s is not None
                for index, reward in enumerate(rewarddict.items()):
                    model.rewards[index].set_state_reward(s, reward[1])

    # we add the observations
    if observations is not None:
        for state, s in state_lookup.items():
            # we check for the observations when it does not return an integer
            given_obs = observations(state)
            if given_obs is None:
                raise ValueError(
                    f"On input {state}, the observations function does not have a return value"
                )

            if isinstance(given_obs, int):
                obs = model.observation(observation_id=given_obs)
                s.set_observation(obs)
            elif isinstance(given_obs, list):
                obs_distribution = [
                    (prob, model.observation(observation_id=o)) for prob, o in given_obs
                ]
                s.set_observation(obs_distribution)
            else:
                raise ValueError(
                    f"On input {state}, the observations function does not return an integer or a distribution"
                )

        if observation_valuations is not None and model.observations is not None:
            observation_valuation_keys = observation_valuations(
                model.observations[0].observation
            ).keys()
            for obs in model.observations:
                valuation_dict = observation_valuations(obs.observation)
                if valuation_dict is None:
                    raise ValueError(
                        f"On input observation id {obs.observation}, the observation_valuations function does not have a return value"
                    )

                if not isinstance(valuation_dict, dict):
                    raise ValueError(
                        f"On input observation id {obs.observation}, the observation_valuations function does not return a dictionary. Make sure to change the format to [<variable>: <value>,...]"
                    )

                if valuation_dict.keys() != observation_valuation_keys:
                    raise RuntimeError(
                        "Make sure that you have a value for each variable in each observation valuation"
                    )

                for val in valuation_dict.values():
                    if not (
                        isinstance(val, int)
                        or isinstance(val, bool)
                        or isinstance(val, float)
                    ):
                        raise ValueError(
                            f"On input observation id {obs.observation}, the dictionary that the observation_valuations function returns contains a value {val} which is not of type int, float or bool"
                        )

                obs.valuations = valuation_dict

    # we add the exit rates
    if rates is not None:
        for state, s in state_lookup.items():
            r = rates(state)
            if not isinstance(r, stormvogel.model.Value):
                raise ValueError(
                    f"On input {state}, the rates function does not return a number"
                )
            model.set_rate(s, r)

    # we add the valuations
    if valuations is not None:
        initial_state_valuations = valuations(init)
        for state, s in state_lookup.items():
            valuation_dict = valuations(state)
            if valuation_dict is None:
                raise ValueError(
                    f"On input {state}, the valuations function does not have a return value"
                )

            if not isinstance(valuation_dict, dict):
                raise ValueError(
                    f"On input {state}, the valuations function does not return a dictionary. Make sure to change the format to [<variable>: <value>,...]"
                )

            if valuation_dict.keys() != initial_state_valuations.keys():
                raise RuntimeError(
                    "Make sure that you have a value for each variable in each state"
                )

            for val in valuation_dict.values():
                if not (
                    isinstance(val, int)
                    or isinstance(val, bool)
                    or isinstance(val, float)
                ):
                    raise ValueError(
                        f"On input {state}, the dictionary that the valuations function returns contains a value {val} which is not of type int, float or bool"
                    )

            s.valuations = valuation_dict

    # we add the labels
    if labels is not None:
        for state, s in state_lookup.items():
            labellist = labels(state)

            assert s is not None

            # if the labels function has no return value we assume this state simply has no labels
            if labellist is None:
                continue

            # we check for the labels when the function does not return a list object, or when
            # the list does not consist of strings
            if not isinstance(labellist, list):
                if not isinstance(labellist, str):
                    raise ValueError(
                        f"On input {state}, the labels function does not return a string or a list of strings"
                    )
                # if we don't get a list, we assume there is just one label
                if labellist not in s.labels:
                    s.add_label(labellist)
            else:
                for label in labellist:
                    if not isinstance(label, str):
                        raise ValueError(
                            f"On input {state}, the labels function does not return a string or a list of strings"
                        )
                    if label not in s.labels:
                        s.add_label(label)

    return model



import numpy as np


def example_building_mas_01():
    import stormpy

    # Building the transition matrix using numpy
    transitions = np.array(
        [
            [0, 1, 0, 0, 0],
            [0.8, 0, 0.2, 0, 0],
            [0.9, 0, 0, 0.1, 0],
            [0, 0, 0, 0, 1],
            [0, 0, 0, 1, 0],
            [0, 0, 0, 0, 1],
        ],
        dtype="float64",
    )

    # Build matrix and define indices of row groups (ascending order)
    transition_matrix = stormpy.build_sparse_matrix(transitions, [0, 2, 3, 4, 5])
    # print(transition_matrix)

    # StateLabeling
    state_labeling = stormpy.storage.StateLabeling(5)
    # Add labels
    state_labels = {"init", "deadlock"}
    # Set labeling of states
    for label in state_labels:
        state_labeling.add_label(label)
    state_labeling.add_label_to_state("init", 0)

    # Choice labeling
    choice_labeling = stormpy.storage.ChoiceLabeling(6)
    # Add labels
    choice_labels = {"alpha", "beta"}
    # Set labeling of choices
    for label in choice_labels:
        choice_labeling.add_label(label)
    choice_labeling.add_label_to_choice("alpha", 0)
    choice_labeling.add_label_to_choice("beta", 1)

    # Provide exit rates (for Markovian states)
    exit_rates = [0.0, 10.0, 12.0, 1.0, 1.0]

    markovian_states = stormpy.BitVector(5, [1, 2, 3, 4])

    # Collect components
    components = stormpy.SparseModelComponents(
        transition_matrix=transition_matrix,
        state_labeling=state_labeling,
        markovian_states=markovian_states,
    )
    components.choice_labeling = choice_labeling
    components.exit_rates = exit_rates

    # Build the model
    ma = stormpy.storage.SparseMA(components)
    return ma


if __name__ == "__main__":
    print(example_building_mas_01())



from typing import Callable, Any
import gymnasium as gym
from collections import defaultdict

from typing import Tuple
from stormvogel import bird
import stormvogel.model


def sample_gym(
    env: gym.Env,
    no_samples: int = 10,
    sample_length: int = 20,
    gymnasium_scheduler: Callable[[Any], int] | None = None,
    convert_obs: Callable[[Any], Any] = lambda x: x,
) -> Tuple:
    """Sample the gym environment. In reality, gym environments are POMDPs, and gymnasium only allows us to access the observation.
    States that are different in gym, but have the same observation and termination will be considered the same state in the result.

    Args:
        env (gym.Env): Gymnasium env.
        no_samples (int): Total number of samples (starting at an initial state).
            To resolve multiple initial states, a new, single initial state is added if necessary.
        sample_length (int): The maximum length of a single sample.
        gymnasium_scheduler (Callable[[any], int] | None): A function from states to action numbers.
        convert_obs (Callable[[any], any]): Converts the observations to a hashable type. You can also apply rounding here.

    Returns:
        A 6-tuple consiting of four defaultdicts and one integer.
        * initial_states (defaultdict[state, int]): The initial state in gym may be non-deterministic. This maps the initial states to the amount of times they were observed as the initial state.
        * states (defaultdict[state, int]): Maps states to the amount of times they were observed.
        * transition_counts (defaultdict[(state,action), defaultdict[state, int]]): Counts how many times the transition between this state-action pair and state was observed.
        * transition_samples (defaultdict[(state,action), int]): Counts how many times this state-action pair was observed.
        * reward_sums (defaultdict[(state,action), int]): The sum of the rewards for this state-action pair.
        * no_actions (int): The number of different actions observed.
    """
    initial_states = defaultdict(lambda: 0)
    states = defaultdict(lambda: 0)
    transition_counts = defaultdict(lambda: defaultdict(lambda: 0))
    transition_samples = defaultdict(lambda: 0)
    reward_sums = defaultdict(lambda: 0.0)

    for s_no in range(no_samples):
        prev_state = None
        obs, _ = env.reset()
        state = (convert_obs(obs), False)
        initial_states[state] += 1
        states[state] += 1
        for _ in range(sample_length):
            action = (
                env.action_space.sample()
                if gymnasium_scheduler is None
                else gymnasium_scheduler(state)
            )
            prev_state = state
            obs, reward, terminated, truncated, info = env.step(action)
            state = (convert_obs(obs), terminated)
            states[state] += 1
            transition_counts[(prev_state, action)][state] += 1
            transition_samples[(prev_state, action)] += 1
            reward_sums[(prev_state, action)] += float(reward)
            if terminated:
                break
    return (
        initial_states,
        states,
        transition_counts,
        transition_samples,
        reward_sums,
        env.action_space.n,
    )


def sample_to_stormvogel(
    initial_states: defaultdict[Any, int],
    transition_counts: defaultdict[Tuple[Any, Any], defaultdict[Any, int]],
    transition_samples: defaultdict[Tuple[Any, Any], int],
    reward_sums: defaultdict[Tuple[Any, Any], int],
    no_actions: int,
    no_samples: int,
    max_size: int = 10000,
) -> stormvogel.model.Model:
    """Create a Stormvogel mdp from a sampling (see sample_gym to obtain a sample from gym).
    Probablities are frequentist estimates. Their accuracy depends on how often each "state" is visited.

    Args:
        initial_states (defaultdict[state, int]): The initial state in gym may be non-deterministic.
            This maps the initial states to the amount of times they were observed as the initial state.
        transition_counts (defaultdict[(state,action), defaultdict[state, int]]):
            Counts how many times the transition between this state-action pair and state was observed.
        transition_samples (defaultdict[(state,action), int]): Counts how many times this state-action pair was observed.
        reward_sums (defaultdict[(state,action), int]): The sum of the rewards for this state-action pair.
        no_actions (int): The number of different actions observed.
        no_samples (int): The number of samples that were used to obtain this sampling.
        max_size (int): The maximum number of states in the resulting model. Defaults to 10000.
    """
    NEW_INITIAL_STATE = "GYM_SAMPLE_INIT"
    ALL_ACTIONS = [str(x) for x in range(no_actions)]
    INV_MAP = {a: no for no, a in enumerate(ALL_ACTIONS)}

    if len(initial_states) == 1:
        (init,) = initial_states
    else:
        init = NEW_INITIAL_STATE

    def available_actions(s):
        if s is NEW_INITIAL_STATE:
            return [""]
        elif s[1]:
            return [""]
        return [a for a in ALL_ACTIONS if transition_counts[(s, INV_MAP[a])]]

    def delta(s, a):
        if s is NEW_INITIAL_STATE:
            return [(count / no_samples, s_) for s_, count in initial_states.items()]
        elif s[1]:
            return [(1, s)]
        return [
            (count / transition_samples[(s, INV_MAP[a])], s_)
            for s_, count in transition_counts[(s, INV_MAP[a])].items()
        ]

    def rewards(s, a) -> dict[str, stormvogel.model.Value]:
        if s is NEW_INITIAL_STATE or s[1]:
            return {"R": 0}
        return {"R": reward_sums[s, INV_MAP[a]] / transition_samples[(s, INV_MAP[a])]}

    def labels(s):
        if s is NEW_INITIAL_STATE:
            return []
        done = ["done"] if s[1] else []
        return [str(s[0])] + done

    return bird.build_bird(
        delta=delta,
        init=init,
        available_actions=available_actions,
        labels=labels,
        rewards=rewards,  # type: ignore
        modeltype=stormvogel.model.ModelType.MDP,
        max_size=max_size,
    )


def sample_gym_to_stormvogel(
    env: gym.Env,
    no_samples: int = 10,
    sample_length: int = 20,
    gymnasium_scheduler: Callable[[Any], int] | None = None,
    convert_obs: Callable[[Any], Any] = lambda x: x,
    max_size: int = 10000,
):
    """Sample the gym environment and convert it to a Stormvogel MDP.
    In reality, gym environments are POMDPs, and gymnasium only allows us to access the observation.
    The result is an MDP where states with the same observations (and termination) are lumped together.
    Probablities are frequentist estimates. Their accuracy depends on how often each "state" is visited.

    Args:
        env (gym.Env): Gymnasium env.
        no_samples (int): Total number of samples (starting at an initial state).
            To resolve multiple initial states, a new, single initial state is added if necessary.
        sample_length (int): The maximum length of a single sample.
        gymnasium_scheduler (Callable[[any], int] | None): A function from states to action numbers.
        convert_obs (Callable[[any], any]): Converts the observations to a hashable type. You can also apply rounding here.
        max_size (int): The maximum number of states in the resulting model. Defaults to 10000.
    """
    (
        initial_states,
        _,
        transition_counts,
        transition_samples,
        reward_sums,
        no_actions,
    ) = sample_gym(env, no_samples, sample_length, gymnasium_scheduler, convert_obs)
    return sample_to_stormvogel(
        initial_states,
        transition_counts,
        transition_samples,
        reward_sums,
        no_actions,
        no_samples,
        max_size,
    )



"""Generate an interactive editor from a specified schema using ipywidgets.
This is used to create the layout editing menu.
Exmample usage can be found in property_builder.py."""

import stormvogel.displayable
import stormvogel.rdict

import ipywidgets as widgets
import IPython.display as ipd
from typing import Any, Callable
import copy


class WidgetWrapper:
    """Creates a widget specified in the arguments."""

    convert_dict = {
        "IntSlider": widgets.IntSlider,
        "FloatSlider": widgets.FloatSlider,
        "ColorPicker": widgets.ColorPicker,
        "Checkbox": widgets.Checkbox,
        "Text": widgets.Text,
        "Dropdown": widgets.Dropdown,
        "TagsInput": widgets.TagsInput,
        "IntText": widgets.IntText,
        "ToggleButtons": widgets.ToggleButtons,
    }

    def __init__(
        self,
        description: str,
        widget: str,
        path: list[str],
        initial_value: Any,
        update_dict: dict,
        on_update: Callable[[], None],
        **kwargs,
    ):
        """Creates a widget which automatically updates a dictonary value.

        Args:
            description (str): 'description' of the widget. Will be shown in the UI.
            widget (str): A str name for a widget.
            path (list[str]): path to the value in update_dict to be changed if the user changes this widget.
            initial_value (Any): initial value of the widget (aka. 'value')
            update_dict (dict): The dict that should be updated.
                When the user sets the value of the widget to X, the value at update_dict[path] is set to X.
            on_update (Callable[[], None]): A function that is called whenever a value is updated.
                When the user sets the value of the widget to something, on_update() is called.
        """
        self.update_dict: dict = update_dict
        self.path: list[str] = path
        self.on_update: Callable = on_update
        # The button is a special case with a slightly different API.
        if widget == "Button":
            self.widget = widgets.Button(description=description, **kwargs)
            self.widget.on_click(self.on_edit)
        else:
            w = self.convert_dict[widget]
            self.widget = widgets.interactive(
                self.on_edit,
                x=w(value=initial_value, description=description, **kwargs),
            )

    def on_edit(self, x: Any) -> None:
        """This method is called when the user sets the value of a widget to x.
        It updates self.update_dict[self.path], and calls the on_update function."""
        stormvogel.rdict.rset(self.update_dict, self.path, x)
        self.on_update()


class DictEditor(stormvogel.displayable.Displayable):
    """Create an interactive json editor from a schema using ipy widgets."""

    def __init__(
        self,
        schema: dict,
        update_dict: dict,
        on_update: Callable[[], None],
        output: widgets.Output | None = None,
        do_display: bool = True,
        debug_output: widgets.Output = widgets.Output(),
    ) -> None:
        """Create an interactive json editor from a schema using ipy widgets. Display it using the show() method.

        Args:
            schema (dict): The dict that specifies the schema.
                Quite closely follows the structure of the actual dict. Supports macros.
                Macros are defined as a dict with the key "__macros" and the value is the name of the macro.
                A macro can be used in the schema by using the key "__use_macro" and the value is the name of the macro.
                See layouts/schema.json for an example.
            update_dict (dict): The dict that will be updated.
            on_update (Callable[[], None]): Function that is called whenever the dict is updated.
            output (widgets.Output | None, optional): The output widget to use.
            do_display (bool, optional): Whether to display the widget immediately.
            debug_output (widgets.Output, optional): Output widget for debug information.
        """
        super().__init__(output, do_display, debug_output)
        self.on_update: Callable = on_update
        self.update_dict: dict = update_dict
        self.macros: dict = {}
        self.schema: dict = schema

    def show(self):
        with self.output:
            ipd.clear_output()
            ipd.display(self.recurse_create(self.schema, []))
        self.maybe_display_output()

    def recurse_create(self, sub_schema: dict, path: list) -> widgets.Widget:
        acc_items = []
        for k, v in sub_schema.items():
            new_path = copy.deepcopy(path)
            new_path.append(k)
            if k == "__html":
                acc_items.append(widgets.HTML(v))
            if k == "__macros":
                self.macros = v
            elif isinstance(v, dict):
                if "__html" in v and "__widget" in v:
                    acc_items.append(widgets.HTML(v["__html"]))
                if "__use_macro" in v:
                    macro_value = self.macros[v["__use_macro"]]
                    acc_items.append(self.recurse_create(macro_value, new_path))
                elif (
                    "__description" in v
                ):  # v is a widget, because it has a defined __description.
                    if "__kwargs" in v:  # Also pass arguments if relevant.
                        w = WidgetWrapper(
                            description=v["__description"],
                            widget=v["__widget"],
                            initial_value=stormvogel.rdict.rget(
                                self.update_dict, new_path
                            ),
                            path=new_path,
                            update_dict=self.update_dict,
                            on_update=self.on_update,
                            **v["__kwargs"],
                        )
                    else:
                        w = WidgetWrapper(
                            description=v["__description"],
                            widget=v["__widget"],
                            initial_value=stormvogel.rdict.rget(
                                self.update_dict, new_path
                            ),
                            path=new_path,
                            update_dict=self.update_dict,
                            on_update=self.on_update,
                        )
                    acc_items.append(w.widget)
                else:  # v is not a widget or macro, then call recursively.
                    acc_items.append(self.recurse_create(v, new_path))
        if "__collapse" in sub_schema and sub_schema["__collapse"]:
            acc = widgets.Accordion(
                children=[widgets.VBox(children=acc_items)], titles=[path[-1]]
            )
        else:
            acc = widgets.VBox(children=acc_items)
        return acc



from stormvogel.model import Model, ModelType


def create_lion_mdp():
    lion = Model(model_type=ModelType.MDP)
    init = lion.get_initial_state()
    full = lion.new_state("full")
    satisfied = init
    hungry = lion.new_state("hungry :(")
    starving = lion.new_state("starving :((")
    dead = lion.new_state("dead")

    hunt = lion.new_action("hunt >:D")
    rawr = lion.new_action("rawr")

    init.set_choice([(1, satisfied)])

    full.set_choice(
        {
            hunt: [(0.5, satisfied), (0.5, full)],
            rawr: [(0.9, full), (0.1, satisfied)],
        }
    )

    satisfied.set_choice(
        {
            hunt: [(0.5, satisfied), (0.3, full), (0.2, hungry)],
            rawr: [(0.9, satisfied), (0.1, hungry)],
        }
    )

    hungry.set_choice(
        {
            hunt: [(0.2, full), (0.5, satisfied), (0.1, hungry), (0.2, starving)],
            rawr: [(0.9, hungry), (0.1, starving)],
        }
    )

    starving.set_choice(
        {
            hunt: [(0.1, full), (0.5, satisfied), (0.2, hungry), (0.2, dead)],
            rawr: [(0.9, starving), (0.1, dead)],
        }
    )

    lion.add_self_loops()

    reward_model = lion.new_reward_model("R")
    reward_model.set_state_action_reward(full, rawr, 100)
    reward_model.set_unset_rewards(0)
    return lion



# Knuth's model of a fair die using only fair coins
def example_building_mdps_01():
    import stormpy

    nr_states = 13
    nr_choices = 14

    # Transition matrix with custom row grouping: nondeterministic choice over the actions available in states
    builder = stormpy.SparseMatrixBuilder(
        rows=0,
        columns=0,
        entries=0,
        force_dimensions=False,
        has_custom_row_grouping=True,
        row_groups=0,
    )

    # New row group, for actions of state 0
    builder.new_row_group(0)
    builder.add_next_value(0, 1, 0.5)
    builder.add_next_value(0, 2, 0.5)
    builder.add_next_value(1, 1, 0.2)
    builder.add_next_value(1, 2, 0.8)
    # State 1
    builder.new_row_group(2)
    builder.add_next_value(2, 3, 0.5)
    builder.add_next_value(2, 4, 0.5)
    # State 2
    builder.new_row_group(3)
    builder.add_next_value(3, 5, 0.5)
    builder.add_next_value(3, 6, 0.5)
    # State 3
    builder.new_row_group(4)
    builder.add_next_value(4, 7, 0.5)
    builder.add_next_value(4, 1, 0.5)
    # State 4
    builder.new_row_group(5)
    builder.add_next_value(5, 8, 0.5)
    builder.add_next_value(5, 9, 0.5)
    # State 5
    builder.new_row_group(6)
    builder.add_next_value(6, 10, 0.5)
    builder.add_next_value(6, 11, 0.5)
    # State 6
    builder.new_row_group(7)
    builder.add_next_value(7, 2, 0.5)
    builder.add_next_value(7, 12, 0.5)

    # Add transitions for the final states
    for s in range(8, 14):
        builder.new_row_group(s)
        builder.add_next_value(s, s - 1, 1)

    transition_matrix = builder.build()

    # State labeling
    state_labeling = stormpy.storage.StateLabeling(nr_states)
    # Add labels
    labels = {"init", "one", "two", "three", "four", "five", "six", "done", "deadlock"}
    for label in labels:
        state_labeling.add_label(label)

    # Set labeling of states
    state_labeling.add_label_to_state("init", 0)
    state_labeling.add_label_to_state("one", 7)
    state_labeling.add_label_to_state("two", 8)
    state_labeling.add_label_to_state("three", 9)
    state_labeling.add_label_to_state("four", 10)
    state_labeling.add_label_to_state("five", 11)
    state_labeling.add_label_to_state("six", 12)

    # Set label 'done' for multiple states
    state_labeling.set_states(
        "done", stormpy.BitVector(nr_states, [7, 8, 9, 10, 11, 12])
    )

    # Choice labeling
    choice_labeling = stormpy.storage.ChoiceLabeling(nr_choices)
    choice_labels = {"a", "b"}
    # Add labels
    for label in choice_labels:
        choice_labeling.add_label(label)

    # Set labels
    choice_labeling.add_label_to_choice("a", 0)
    choice_labeling.add_label_to_choice("b", 1)
    # print(choice_labeling)

    # Reward models
    reward_models = {}
    # Create a vector representing the state-action rewards
    action_reward = [
        0.0,
        0.0,
        1.0,
        1.0,
        1.0,
        1.0,
        1.0,
        1.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
    ]
    reward_models["coin_flips"] = stormpy.SparseRewardModel(
        optional_state_action_reward_vector=action_reward
    )

    # Collect components
    components = stormpy.SparseModelComponents(
        transition_matrix=transition_matrix,
        state_labeling=state_labeling,
        reward_models=reward_models,
        rate_transitions=False,
    )
    components.choice_labeling = choice_labeling

    # Build the model
    mdp = stormpy.storage.SparseMdp(components)
    return mdp


if __name__ == "__main__":
    example_building_mdps_01()



from stormvogel.stormpy_utils.convert_results import *  # NOQA
from stormvogel.stormpy_utils.mapping import *  # NOQA

# Note that we do not import magic here! This is done in the __init__.py of stormvogel itself.
